""" New/Edit dialog (dialog is blocking the main-window, as opposed to create a new widget-window)"""
import copy
import json
import logging
import re
import warnings
from enum import Enum
from pathlib import Path
from typing import Any, Union
import pandas as pd
from PySide6.QtCore import QSize, Qt, QTimer, Slot
from PySide6.QtGui import QRegularExpressionValidator
from PySide6.QtWidgets import (QComboBox, QDialog, QFormLayout, QHBoxLayout, QLabel, QLayout, QLineEdit, QMessageBox,
                               QScrollArea, QSizePolicy, QSplitter, QTabWidget, QTextEdit, QVBoxLayout, QWidget)
from ..backendWorker.sqlite import MAIN_ORDER
from ..backendWorker.worker import Task
from ..fixedStringsJson import SQLiteTranslationDict, defaultDataHierarchyNode, minimalDocInForm
from ..miscTools import callAddOn
from ..textTools.stringChanges import markdownEqualizer
from ._contextMenu import CommandMenu, executeContextMenu, initContextMenu
from .guiCommunicate import Communicate
from .guiStyle import IconButton, Image, Label, ScrollMessageBox, TextButton, widgetAndLayout, widgetAndLayoutForm
from .messageDialog import showMessage
from .textEditor import TextEditor


class Form(QDialog):
  """ New/Edit dialog (dialog is blocking the main-window, as opposed to create a new widget-window)"""
  def __init__(self, comm:Communicate, doc:dict[str,Any]):
    """
    Initialization

    Args:
      comm (Communicate): communication channel
      doc (dict):  document to change / create
    """
    super().__init__()
    self.comm = comm

    # get data into shape
    self.doc:dict[str,Any] = copy.deepcopy(doc)
    self.allDocIDs = self.doc.get('_ids', [self.doc['id']] if 'id' in self.doc else [])
    self.allDocIDsCopy = list(self.allDocIDs)
    self.flagNewDoc = 'id' not in self.doc and '_ids' not in self.doc
    self.groupEdit = '_ids' in self.doc
    self.allowDocTypeChange = self.doc.get('id',' ')[0] != 'x'
    if len(self.allDocIDs)>1:                                                                   #if group edit
      self.allowDocTypeChange = all(docID[0] != 'x' for docID in self.allDocIDs)
    self.allowDocTypeChange = self.allowDocTypeChange and not self.flagNewDoc
    self.allowProjectChange = True

    # GUI elements
    if self.flagNewDoc:
      self.setWindowTitle('Create new entry')
    else:
      self.setWindowTitle('Edit information')
    self.setMinimumSize(QSize(800,600))
    self.mainL = QVBoxLayout(self)
    self.splitter = QSplitter(Qt.Orientation.Horizontal)                         # will be filled during paint
    self.splitter.setHandleWidth(10)
    self.splitter.setContentsMargins(0,0,0,0)
    self.mainL.addWidget(self.splitter, stretch=2)

    # final button box
    _, buttonLineL = widgetAndLayout('H', self.mainL, 'm')
    self.visibilityText = QLabel('')
    buttonLineL.addWidget(self.visibilityText)
    buttonLineL.addStretch(1)
    self.btnAddKWPairs = IconButton('ri.menu-add-fill', self, [Command.FORM_ADD_KV],   buttonLineL,
                                    'Add key-value pair', style='border-width:1')
    self.btnAddKWPairs.setFocusPolicy(Qt.FocusPolicy.NoFocus)
    self.btnAddKWPairs.setAutoDefault(False)
    self.btnAddKWPairs.setDefault(False)
    if not self.flagNewDoc:                                                                  #existing dataset
      self.showDocBtn = IconButton('fa5s.poll-h',      self, [Command.FORM_SHOW_DOC], buttonLineL, 'Show all information',
                 style='border-width:1')
      self.showDocBtn.setFocusPolicy(Qt.FocusPolicy.NoFocus)
      self.showDocBtn.setAutoDefault(False)
      self.btnDuplicate = IconButton('fa5s.plus-circle', self, [Command.FORM_SAVE_NEXT], buttonLineL,
                                     'Duplicate data set', style='border-width:1')
      self.btnDuplicate.setFocusPolicy(Qt.FocusPolicy.NoFocus)
      self.btnDuplicate.setAutoDefault(False)
    self.saveBtn = TextButton('Save',             self, [Command.FORM_SAVE],     buttonLineL, 'Save changes')
    self.saveBtn.setShortcut('Ctrl+Return')
    self.saveBtn.setFocusPolicy(Qt.FocusPolicy.NoFocus)
    self.saveBtn.setAutoDefault(False)
    self.cancelBtn = TextButton('Cancel',           self, [Command.FORM_CANCEL],   buttonLineL, 'Discard changes')
    self.cancelBtn.setFocusPolicy(Qt.FocusPolicy.NoFocus)
    self.cancelBtn.setAutoDefault(False)
    if self.flagNewDoc:                                                                           #new dataset
      self.saveNextBtn = TextButton('Save && Next', self, [Command.FORM_SAVE_NEXT], buttonLineL, 'Save this and handle next')
      self.saveNextBtn.setFocusPolicy(Qt.FocusPolicy.NoFocus)
      self.saveNextBtn.setAutoDefault(False)
    self.setStyleSheet(f"QLineEdit, QComboBox {{ {self.comm.palette.get('secondaryText', 'color')} }}")

    #GUI elements, filled later
    self.tagsBarMainW                         = QWidget()
    self.gradeChoices                         = QComboBox()
    self.tagsBarSubL:QLayout|None             = None
    self.otherChoices                         = QComboBox()
    self.keyValueListW                        = QWidget()
    self.keyValueListL:QFormLayout|None       = None
    self.keyValueLabel                        = QLabel()
    self.projectComboBox                      = QComboBox()
    self.skipKeys  = ['image','metaVendor','metaUser','shasum','._projectID','._ids','.name','.elnIdentifier']
    self.allHidden                            = False
    self.keyLabels: list[QLineEdit]           = []
    self.values:    list[QLineEdit]           = []
    self.imageL:QLayout|None                  = None
    self.dataHierarchyNode:list[dict[str,str]]= [{}]
    self.tabW                                 = QTabWidget()                    # has count=0 if not connected
    self.formsL:list[QLayout]                 = []
    self.allUserElements:list[tuple[str,str]] = []
    self.docTypeComboBox                      = QComboBox()
    self.checkThreadTimer                     = QTimer(self)
    self.tagsAllList: list[str] = []
    self.comboBoxDocTypeList:dict[str, tuple[QComboBox,str]] = {}# dict of docType:.. for links to other items

    # Request data
    if not self.flagNewDoc:
      self.comm.backendThread.worker.beSendDoc.connect(self.onGetData)    # do not wait for data if not needed
    self.comm.backendThread.worker.beSendTable.connect(self.onGetTable)          # tags, docTypes to link, ...
    for docID in self.allDocIDs:
      self.comm.uiRequestDoc.emit(docID)
    self.comm.uiRequestTable.emit('_tags_','', True)
    self.comm.uiRequestTable.emit('x0','', True)

    # setup Autosave, start it and paint form
    if (Path.home()/'.pastaELN.temp').is_file():
      with open(Path.home()/'.pastaELN.temp', encoding='utf-8') as fTemp:
        content = json.loads(fTemp.read())
        if self.doc.get('id', '') in content:
          ret = QMessageBox.information(self, 'Information', 'There is unsaved information from a prematurely '+
                    'closed form. Do you want to restore it?\n If you decline, the unsaved information will be'+
                    ' removed.',
                  QMessageBox.StandardButton.No | QMessageBox.StandardButton.Yes,
                  QMessageBox.StandardButton.Yes)
          if ret==QMessageBox.StandardButton.Yes:
            subContent = content[self.doc.get('id', '')]
            for key in subContent.keys():
              try:
                elementName = f'key_{[idx for idx, (k,_) in enumerate(self.allUserElements) if key==k][0]}'
              except Exception:
                continue
              if key in ('comment', 'content'):
                getattr(self, f'textEdit_{key}').setPlainText(subContent[key])
              elif key in ('tags'):
                self.doc[key] = subContent[key]
                self.updateTagsBar()
              elif isinstance(getattr(self, elementName), QLineEdit):
                getattr(self, elementName).setText(subContent[key])
              # skip QCombobox items since cannot be sure that next from has them and they are easy to recreate
          del content[self.doc.get('id', '')]
      with open(Path.home()/'.pastaELN.temp', 'w', encoding='utf-8') as fTemp:
        fTemp.write(json.dumps(content))
    self.checkThreadTimer.setInterval(1*60*1000)                                                       # 1 min
    self.checkThreadTimer.timeout.connect(self.autosave)
    self.checkThreadTimer.start()
    self.paint()


  @Slot(dict)
  def onGetData(self, doc:dict[str,Any]) -> None:
    """
    Get data from the backend

    Args:
      doc (dict):  document to change / create
    """
    if not doc:
      return
    if doc['id'] in self.allDocIDs:
      self.allDocIDs.remove(doc['id'])
      if 'id' in self.doc and len(self.doc)==1:             # initialize self.doc with a real doc: not just id
        self.doc = doc
      else:
        intersection = set(self.doc).intersection(set(doc))
        #remove keys that should not be group edited and build dict
        intersection = intersection.difference({'branch', 'user', 'client', 'metaVendor', 'shasum', 'id', 'type',
                           'metaUser', 'rev', 'name', 'dateCreated', 'dateModified', 'image', 'gui', ''})
        docType = list(self.doc['type'])
        self.doc = {i:'' for i in intersection}
        self.doc['tags'] = []
        self.doc['type'] = docType
      self.allowProjectChange = self.allowProjectChange and self.doc['type'][0]!='x0'# none of the items can be a project
    if len(self.allDocIDs)==0:
      self.paint()



  @Slot(pd.DataFrame, str)
  def onGetTable(self, data:pd.DataFrame, docType:str) -> None:
    """
    - Get tags from the backend
    - get list of projects and other docTypes and fill the comboBoxes

    Args:
      data (pd.DataFrame):  DataFrame containing tags
      docType (str): document type
    """
    if docType == '_tags_':
      self.tagsAllList = data['tag'].unique()
      self.updateTagsBar()
    elif docType == 'x0':
      self.projectComboBox.clear()
      self.projectComboBox.addItem('- no change' if self.groupEdit else '- not assigned -', userData='')
      for iDocID, iName in data[['id','name']].values.tolist():           # add all projects incl. the present
        self.projectComboBox.addItem(iName, userData=iDocID)
        stack = self.doc.get('branch',[{}])[0].get('stack', [])
        proj  = stack[0] if stack else ''
        if iDocID in (self.doc.get('_projectID',''), proj):
          self.projectComboBox.setCurrentIndex(self.projectComboBox.count()-1)
    elif docType in self.comboBoxDocTypeList:
      iComboBox, value = self.comboBoxDocTypeList[docType]
      iComboBox.clear()
      iComboBox.addItem('- no link -', userData='')
      for iDocID, iName in data[['id','name']].values.tolist():
        iComboBox.addItem(iName, userData=iDocID)
        if iDocID == value:
          iComboBox.setCurrentIndex(iComboBox.count()-1)
    else:
      logging.warning('Unknown docType in onGetTable: %s', docType)


  def paint(self) -> None:
    """ Paint the form with all the elements """
    if 'type' not in self.doc:
      return
    if '_attachments' in self.doc:
      del self.doc['_attachments']
    for i in reversed(range(self.splitter.count())):                        # remove all widgets from splitter
      widget = self.splitter.widget(i)
      if widget is not None and widget is not self.projectComboBox:
        widget.setParent(None)
    self.comboBoxDocTypeList = {}                                                  # reset comboBoxDocTypeList
    self.allUserElements     = []
    self.doc = copy.deepcopy(minimalDocInForm) | self.doc
    if self.flagNewDoc:
      self.doc['name'] = ''
    if 'branch' in self.doc:
      visibilityIcon = all(all(branch['show']) for branch in self.doc['branch'])
      self.visibilityText = QLabel('' if visibilityIcon else 'HIDDEN     \U0001F441')
      self.btnDuplicate.setHidden(self.doc['branch'][0]['path'] is not None)

    # image
    if 'image' in self.doc:
      imageWSA = QScrollArea()
      imageWSA.setWidgetResizable(True)
      imageW, self.imageL = widgetAndLayout('V', self.splitter)
      width= self.comm.configuration['GUI']['imageSizeDetails'] if hasattr(self.comm,'configuration') else 300
      Image(self.doc['image'], self.imageL, anyDimension=width)
      if 'id' in self.doc:
        imageW.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        imageW.customContextMenuRequested.connect(lambda pos: initContextMenu(self, pos))
      imageWSA.setWidget(imageW)
      self.splitter.addWidget(imageWSA)
      self.setMinimumWidth(1100)
      self.setMinimumHeight(600)
    else:
      self.setMinimumWidth(600)

    # create data hierarchy node: data structure
    if self.doc['type'][0] in self.comm.docTypesTitles:
      rawData = self.comm.dataHierarchyNodes[self.doc['type'][0]]
      self.dataHierarchyNode = copy.deepcopy([dict(i) for i in rawData])
    else:
      self.dataHierarchyNode = copy.deepcopy(defaultDataHierarchyNode)
    keysDataHierarchy = [f"{i['class']}.{i['name']}" for i in self.dataHierarchyNode]
    keysDocOrg = [[str(x) for x in (f'{k}.{k1}' for k1 in self.doc[k])] if isinstance(self.doc[k], dict) else [f'.{k}']
               for k in self.doc if k not in MAIN_ORDER+['branch','qrCodes','tags']]
    for keyInDocNotHierarchy in {i for row in keysDocOrg for i in row}.difference(keysDataHierarchy):
      group = keyInDocNotHierarchy.split('.')[0]
      key = keyInDocNotHierarchy.split('.')[1]
      if not key:
        continue
      idx = len([1 for i in self.dataHierarchyNode if i['class']==group])
      self.dataHierarchyNode.append({'docType': self.doc['type'][0], 'class':group, 'idx':str(idx), 'name':key,
                                     'list':'', 'mandatory':'', 'unit':''})
    #TODO: TEMPORARY CHECK: REMOVE 2026
    allKeys = {'docType', 'class', 'idx', 'name', 'unit', 'mandatory', 'list'}
    if any(allKeys.difference(i.keys()) for i in self.dataHierarchyNode):
      #mask = [allKeys.difference(i.keys()) for i in self.dataHierarchyNode]
      raise ValueError('dataHierarchyNode is not complete. Missing keys')
    # END TEMPORARY CHECK
    groups = {i['class'] for i in self.dataHierarchyNode}.difference({'metaVendor','metaUser'})
    # create tabs or not: depending on the number of groups
    if len(groups)>1:
      self.tabW.setParent(self)
      self.tabW.tabBarClicked.connect(self.changeTabs)
      self.splitter.addWidget(self.tabW)

    # create forms by looping
    for group in groups:
      if len(groups)==1:
        _, formL = widgetAndLayoutForm(self.splitter, 's')
      else:
        formW, formL = widgetAndLayoutForm(None, 's', top='m')
        self.tabW.addTab(formW, 'Home' if group=='' else group)
      self.formsL.append(formL)
      for name in [i['name'] for i in self.dataHierarchyNode if i['class']==group]:
        key = f"{group}.{name}"
        defaultValue = self.doc['qrCodes'] if key=='.qrCodes' and 'qrCodes' in self.doc else \
                       self.doc.get(group, {}).get(name, ('','','',''))#tags, name, comment are handled separately
        elementName = f"key_{len(self.allUserElements)}"

        # case list
        if key == '.name' and not self.groupEdit:
          setattr(self, elementName, QLineEdit(self.doc.get('name','')))
          getattr(self, elementName).setValidator(QRegularExpressionValidator('[\\w\\ .-]+'))
          formL.addRow('Name', getattr(self, elementName))
          self.allUserElements.append(('name','LineEdit'))
        elif key == '.tags':
          self.tagsBarMainW, tagsBarMainL = widgetAndLayout('H', spacing='s')
          self.gradeChoices = QComboBox()                                     #part/combobox that shows grades
          self.gradeChoices.setMaximumWidth(95)
          self.gradeChoices.setStyleSheet('padding: 3px;')
          self.gradeChoices.setIconSize(QSize(0,0))
          self.gradeChoices.addItems(['none','\u2605','\u2605'*2,'\u2605'*3,'\u2605'*4,'\u2605'*5])
          gradeTag = [i for i in self.doc['tags'] if i.startswith('_')]
          gradeTagStr = '\u2605'*int(gradeTag[0][1]) if gradeTag else ''
          self.gradeChoices.setCurrentText(gradeTagStr)
          tagsBarMainL.addWidget(self.gradeChoices)
          Label('Tags:', '',  tagsBarMainL, style='margin-left: 20px;')
          tagsBarSubW, self.tagsBarSubL = widgetAndLayout('H', tagsBarMainL, spacing='s', right='m', left='m')#part which shows all the tags
          tagsBarSubW.setMaximumWidth(420)
          self.otherChoices = QComboBox()                             #part/combobox that allow user to select
          self.otherChoices.setToolTip('Choose a tag or type a new one')
          self.otherChoices.setEditable(True)
          self.otherChoices.setMinimumWidth(80)
          self.otherChoices.setValidator(QRegularExpressionValidator('[a-zA-Z]\\w+'))
          self.otherChoices.setIconSize(QSize(0,0))
          self.otherChoices.setInsertPolicy(QComboBox.InsertPolicy.InsertAtBottom)
          self.otherChoices.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
          tagsBarMainL.addWidget(self.otherChoices)
          formL.addRow(QLabel('Rating:'), self.tagsBarMainW)
          self.allUserElements.append(('tags',''))
          self.updateTagsBar()
          self.otherChoices.currentIndexChanged.connect(self.addTag)#connect to slot only after all painting is done
        elif key in ['.comment', '.content']:
          key = key[1:]
          labelW, labelL = widgetAndLayout('V', spacing='s')
          labelL.addWidget(QLabel(key.capitalize()))
          TextButton('More', self, [Command.FOCUS_AREA, key], labelL, checkable=True)
          projectGroup = self.comm.configuration['projectGroups'][self.comm.projectGroup]
          if 'form' in projectGroup.get('addOns',{}) and projectGroup['addOns']['form']:
            TextButton('Auto', self, [Command.AUTO_COMMENT],    labelL, checkable=True)
          rightSideW, rightSideL = widgetAndLayout('V')
          setattr(self, f'buttonBarW_{key}', QWidget())
          getattr(self, f'buttonBarW_{key}').hide()
          buttonBarL = QHBoxLayout(getattr(self, f'buttonBarW_{key}'))
          for name, tooltip in [['bold','Bold text'],['italic','Italic text'],['list-ul','Bullet list'],\
                                      ['list-ol','Numbered list']]:
            IconButton(f'fa5s.{name}', self, [Command.BUTTON_BAR, name, key], buttonBarL, tooltip)
          for i in range(1,4):
            icon = f'mdi.format-header-{str(i)}'
            IconButton(icon, self, [Command.BUTTON_BAR, f'heading{str(i)}', key], buttonBarL, f'Heading {str(i)}')
          rightSideL.addWidget(getattr(self, f'buttonBarW_{key}'))
          textStr = self.doc.get(key, '')
          for k,v in SQLiteTranslationDict.items():
            textStr = textStr.replace(v,k)
          setattr(self, f'textEdit_{key}', TextEditor(textStr))
          getattr(self, f'textEdit_{key}').setAccessibleName(key)
          getattr(self, f'textEdit_{key}').textChanged.connect(self.textChanged)
          setattr(self, f'textShow_{key}', QTextEdit())
          getattr(self, f'textShow_{key}').setMarkdown(markdownEqualizer(self.doc.get(key, '')))
          getattr(self, f'textShow_{key}').setReadOnly(True)
          getattr(self, f'textShow_{key}').hide()
          splittedEditor = QSplitter()
          splittedEditor.setSizePolicy(QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding))
          splittedEditor.addWidget(getattr(self, f'textEdit_{key}'))
          splittedEditor.addWidget(getattr(self, f'textShow_{key}'))
          rightSideL.addWidget(splittedEditor)
          formL.addRow(labelW, rightSideW)
          self.allUserElements.append((key, 'comment'))
        elif key in self.skipKeys:                                                      #skip non desired ones
          continue
        elif isinstance(defaultValue, list):                                 #list of items, qrCodes in sample
          if len(defaultValue)>0 and isinstance(defaultValue[0], str):
            setattr(self, elementName, QLineEdit(' '.join(defaultValue)))
            self.allUserElements.append((key,'LineEdit'))
            formL.addRow(QLabel(key.capitalize()), getattr(self, elementName))
          else:
            logging.info('Cannot display value of key=%s: %s. Write unknown value for docID=%s',
                         key, str(defaultValue), self.doc['id'])
        elif (isinstance(defaultValue, tuple) and len(defaultValue)==4 and isinstance(defaultValue[0], str)) or \
              isinstance(defaultValue, str):                                                  #string or tuple
          dataHierarchyItem = [i for i in self.dataHierarchyNode if i['class']==group and f"{i['class']}.{i['name']}"==key]
          if len(dataHierarchyItem)!=1:
            raise ValueError('more than one dataHierarchyItem')
          label = dataHierarchyItem[0]['name'].capitalize()
          if isinstance(defaultValue, str):
            value = defaultValue
          else:                                                                                         #tuple
            value = defaultValue[0]
            label += '' if defaultValue[1] is None or defaultValue[1]=='' else f' [{defaultValue[1]}]'
            label += '' if defaultValue[3] is None or defaultValue[3]=='' else f'&nbsp;<b><a href="{defaultValue[3]}">&uArr;</a></b>'
          if dataHierarchyItem[0]['list']:                                                    #choice dropdown
            setattr(self, elementName, QComboBox())
            if ',' in dataHierarchyItem[0]['list']:                             #dataHierarchy-defined choices
              getattr(self, elementName).addItems(dataHierarchyItem[0]['list'].split(','))
              getattr(self, elementName).setCurrentText(value)
            else:                                                                        #choice among docType
              listDocType = dataHierarchyItem[0]['list']
              if listDocType not in self.comboBoxDocTypeList:          # if listDocType already exists in dict
                self.comm.uiRequestTable.emit(listDocType, '', True)
                self.comboBoxDocTypeList[listDocType] = (getattr(self, elementName), value)
            self.allUserElements.append((key,'ComboBox'))
          else:                                                                                     #text area
            setattr(self, elementName, QLineEdit(value))
            self.allUserElements.append((key,'LineEdit'))
          formLabelW = QLabel(label)
          formLabelW.setOpenExternalLinks(True)
          formLabelW.setTextInteractionFlags(Qt.TextInteractionFlag.LinksAccessibleByMouse)
          formL.addRow(formLabelW, getattr(self, elementName))
        else:
          print(f"**WARNING dialogForm: unknown value type. key:{key}, type:{type(defaultValue)}")
        if group in self.doc and name in self.doc[group]:
          del self.doc[group][name]
          if not self.doc[group]:
            del self.doc[group]
      if group == '':
        # individual key-value items
        self.keyValueListW, self.keyValueListL = widgetAndLayoutForm(None, 's')
        self.keyValueListW.hide()
        self.keyValueLabel = QLabel('Key - values')
        self.keyValueLabel.hide()
        formL.addRow(self.keyValueLabel, self.keyValueListW)
        ### add extra questions at bottom of form
        # project change
        if self.allowProjectChange:
          formL.addRow(QLabel('Project'), self.projectComboBox)
        # docType change
        if self.allowDocTypeChange:                                                 #if not-new and non-folder
          self.docTypeComboBox = QComboBox()
          self.docTypeComboBox.addItem('- no change -', userData='')
          for key1, value1 in self.comm.docTypesTitles.items():
            if key1[0]!='x':
              self.docTypeComboBox.addItem(value1['title'], userData=key1)
          self.docTypeComboBox.addItem('_UNIDENTIFIED_', userData='-')
          formL.addRow(QLabel('Item type'), self.docTypeComboBox)
    if [i for i in self.doc if i.startswith('_') and i not in ['_projectID']]:
      logging.error('There should not be "_" in a doc: %s', str(self.doc), exc_info=True)


  def autosave(self) -> None:
    """ Autosave comment to file """
    if self.comm.configuration['GUI']['autosave'] == 'No':
      return
    subContent = {'name':getattr(self, f"key_{self.allUserElements.index(('name','LineEdit'))}").text().strip(),
                  'tags':self.doc['tags']}
    for idx, (key, guiType) in enumerate(self.allUserElements):
      if key in ['tags', 'name']:
        continue
      elementName = f"key_{idx}"
      if key in ['comment','content']:
        subContent[key] = getattr(self, f'textEdit_{key}').toPlainText().strip()
      elif key in self.skipKeys:
        continue
      elif guiType=='ComboBox':
        valueNew = getattr(self, elementName).currentText()
        dataNew  = getattr(self, elementName).currentData()                 #if docID is stored in currentData
        if ((dataNew is not None and re.search(r'^[a-z\-]-[a-z0-9]{32}$',dataNew) is not None)
            or dataNew==''):
          subContent[key] = dataNew
        elif valueNew!='- no link -' or dataNew is None:
          subContent[key] = valueNew
      else:                                                                                 #normal text field
        subContent[key] = getattr(self, elementName).text().strip()
    # skip QCombobox items since cannot be sure that next from has them and they are easy to recreate
    if (Path.home()/'.pastaELN.temp').is_file():
      with open(Path.home()/'.pastaELN.temp', encoding='utf-8') as fTemp:
        content = json.loads(fTemp.read())
    else:
      content = {}
    content[self.doc.get('id', '')] = subContent
    with open(Path.home()/'.pastaELN.temp', 'w', encoding='utf-8') as fTemp:
      fTemp.write(json.dumps(content))
    return


  def execute(self, command:list[Any]) -> None:
    """
    Event if user clicks button

    Args:
      command (list): list of commands
    """
    if isinstance(command[0], CommandMenu):
      if executeContextMenu(self, command) and self.imageL is not None:
        item = self.imageL.itemAt(0)
        if item is not None: item.widget().setParent(None)
        width=self.comm.configuration['GUI']['imageSizeDetails'] if hasattr(self.comm,'configuration') else 300
        if 'image' in self.doc:
          Image(self.doc['image'], self.imageL, anyDimension=width)
        if 'branch' in self.doc:
          visibilityIcon = all(all(branch['show']) for branch in self.doc['branch'])
          self.visibilityText.setText('' if visibilityIcon else 'HIDDEN     \U0001F441')

    elif command[0] is Command.BUTTON_BAR:
      if command[1]=='bold':
        getattr(self, f'textEdit_{command[2]}').insertPlainText('**TEXT**')
      elif command[1]=='italic':
        getattr(self, f'textEdit_{command[2]}').insertPlainText('*TEXT*')
      elif command[1]=='list-ul':
        getattr(self, f'textEdit_{command[2]}').insertPlainText('\n- item 1\n- item 2')
      elif command[1]=='list-ol':
        getattr(self, f'textEdit_{command[2]}').insertPlainText('\n1. item 1\n1. item 2')
      elif command[1].startswith('heading'):
        getattr(self, f'textEdit_{command[2]}').insertPlainText('#' * int(command[1][-1]) +' Heading\n')

    elif command[0] is Command.FOCUS_AREA:
      unknownWidget = []
      idx = 0 if self.tabW.count()==0 else self.tabW.currentIndex()
      if self.allHidden:                                       #hide the special buttons and show general form
        getattr(self, f'textShow_{command[1]}').hide()
        getattr(self, f'buttonBarW_{command[1]}').hide()
        for i in range(self.formsL[idx].count()):
          item = self.formsL[idx].itemAt(i)
          widget = None if item is None else item.widget()
          if isinstance(widget, (QLabel, QComboBox, QLineEdit)):
            widget.show()
          else:
            unknownWidget.append(i)
        self.tagsBarMainW.show()
        if command[1]=='content' and len(unknownWidget)==5:#show / hide label and right-side of non-content and non-comment
          item = self.formsL[idx].itemAt(unknownWidget[0])
          if item is not None: item.widget().show()
          item = self.formsL[idx].itemAt(unknownWidget[1])
          if item is not None: item.widget().show()
        if command[1]=='comment' and len(unknownWidget)==5:
          item = self.formsL[idx].itemAt(unknownWidget[2])
          if item is not None: item.widget().show()
          item = self.formsL[idx].itemAt(unknownWidget[3])
          if item is not None:  item.widget().show()
        if self.keyValueListL is not None and self.keyValueListL.count() == 0:
          self.keyValueLabel.hide()
          self.keyValueListW.hide()
      else:                                   #show buttons to allow for easy markdown edit; hide general form
        getattr(self, f'textShow_{command[1]}').show()
        getattr(self, f'buttonBarW_{command[1]}').show()
        for i in range(self.formsL[idx].count()):
          item = self.formsL[idx].itemAt(i)
          widget = None if item is None else item.widget()
          if isinstance(widget, (QLabel, QComboBox, QLineEdit)):
            widget.hide()
          else:
            unknownWidget.append(i)
        self.tagsBarMainW.hide()
        if command[1]=='content' and len(unknownWidget)==5:
          item = self.formsL[idx].itemAt(unknownWidget[0])
          if item is not None: item.widget().hide()
          item = self.formsL[idx].itemAt(unknownWidget[1])
          if item is not None: item.widget().hide()
        if command[1]=='comment' and len(unknownWidget)==5:
          item = self.formsL[idx].itemAt(unknownWidget[2])
          if item is not None: item.widget().hide()
          item = self.formsL[idx].itemAt(unknownWidget[3])
          if item is not None: item.widget().hide()
      self.allHidden = not self.allHidden

    elif command[0] is Command.AUTO_COMMENT:
      try:
        text  = f'\n{"-"*20}\n'
        text += callAddOn('form_auto', self.comm, self.doc, self)
        text += f'\n{"-"*20}\n'
        self.textEdit_comment.insertPlainText(text)                               # type: ignore[attr-defined]
      except Exception:
        pass

    elif command[0] is Command.FORM_CANCEL:
      if self.comm. configuration['GUI']['autosave'] == 'Yes':
        ret = QMessageBox.critical(self, 'Warning', 'You will lose the entered information. Do you want to '+
          'save everything to a temporary location?',
          QMessageBox.StandardButton.Cancel | QMessageBox.StandardButton.No | QMessageBox.StandardButton.Yes,
          QMessageBox.StandardButton.No)
        if ret==QMessageBox.StandardButton.Cancel:
          return
        if ret==QMessageBox.StandardButton.Yes:
          self.autosave()
      self.checkThreadTimer.stop()
      self.reject()

    elif command[0] in (Command.FORM_SAVE, Command.FORM_SAVE_NEXT, Command.FORM_SAVE_DUPL):
      self.checkThreadTimer.stop()
      if (Path.home()/'.pastaELN.temp').is_file():                               #if there is a temporary file
        with open(Path.home()/'.pastaELN.temp', encoding='utf-8') as fTemp:
          content = json.loads(fTemp.read())
          if self.doc.get('id', '') in content:            #if there is an ID in content -> unique identifier?
            del content[self.doc.get('id', '')]                                                # delete the ID
        with open(Path.home()/'.pastaELN.temp', 'w', encoding='utf-8') as fTemp:  #reopen temporary file again
          fTemp.write(json.dumps(content))

      # loop through all the subitems
      # rating
      rating = self.gradeChoices.currentText()
      self.doc['tags'] = [i for i in self.doc['tags'] if not i.startswith('_')]       # filter out old ratings
      if rating != 'none':
        self.doc['tags'].append(f'_{len(rating)}')
      for idx, (key, guiType) in enumerate(self.allUserElements):
        elementName = f"key_{idx}"
        valueOld = self.doc.get(key, '')
        if '.' in key:
          group, subItem = key.split('.')
        else:
          group, subItem = '', key
        if [i['mandatory'] for i in self.dataHierarchyNode if i['class']==group and i['name']==subItem] == ['T'] and \
          getattr(self, elementName).text().strip()=='':
          print(f'**ERROR group:{group}| key:{key}| subItem:{subItem}: mandatory field is empty')
          showMessage(self, 'Error', f'The created item must have a valid {subItem}')
          return
        if key=='name':
          self.doc['name'] = getattr(self, elementName).text().strip()
        elif key in ('tags'):                                                          #tags are already saved
        #     'image', 'metaVendor', 'metaUser' or not hasattr(self, f'key_{key}') and not hasattr(self, f'textEdit_{key}')):
          continue
        elif key == '.qrCodes':
          self.doc['qrCodes'] = getattr(self, elementName).text().strip().split('/')
        elif key in ('comment','content'):
          text = getattr(self, f'textEdit_{key}').toPlainText().strip()
          if '_ids' not in self.doc or text:                             #if group edit, text has to have text
            self.doc[key] = text
            if key == 'content' and 'branch' in self.doc:
              for branch in self.doc['branch']:
                if branch['path'] is not None and branch['path'].endswith('.md'):
                  with open(self.comm.basePath/branch['path'], 'w', encoding='utf-8') as fOut:
                    fOut.write(self.doc['content'])
                  logging.debug('Wrote new content to %s',branch['path'])
                elif branch['path'] is not None:
                  showMessage(self, 'Information', 'Did update the database but not the file on harddisk, since PASTA-ELN cannot write this format')
        elif isinstance(valueOld, list):                     #items that are comma separated in the text-field
          self.doc[key] = getattr(self, elementName).text().strip().split(' ')
        elif isinstance(valueOld, str):
          if guiType=='ComboBox':
            valueNew = getattr(self, elementName).currentText()
            dataNew  = getattr(self, elementName).currentData()             #if docID is stored in currentData
            if ((dataNew is not None and re.search(r'^[a-z\-]-[a-z0-9]{32}$',dataNew) is not None)
                or dataNew==''):
              self.doc[key] = dataNew
            elif valueNew!='- no link -' or dataNew is None:
              self.doc[key] = valueNew
          else:                                                                             #normal text field
            self.doc[key] = getattr(self, elementName).text().strip()
        # elif valueOld is None and key in self.doc:  #important entry, set to empty string
        #   self.doc[key]=''
        #   print('Is this really needed?')
        else:
          logging.error('Unknown value type %s %s',key, valueOld, exc_info=True)
      # new key-value pairs
      keyValueList = [self.keyValueListL.itemAt(i).widget().text() for i in range(self.keyValueListL.count())]# type: ignore[union-attr]
      keyValueDict = dict(zip(keyValueList[::2],keyValueList[1::2] ))
      keyValueDict = {f'.{k}':v for k,v in keyValueDict.items() if k}
      self.doc = keyValueDict | self.doc
      # ---- if project changed: only branch save; remaining data still needs saving
      newProjID = [self.projectComboBox.currentData()] if self.projectComboBox.currentData() else []
      # ---- if docType changed: save; no further save to db required ----
      if self.docTypeComboBox.currentData() != '' and self.docTypeComboBox.currentData() is not None:
        self.doc['type'] = [self.docTypeComboBox.currentData()]

      docBackup = copy.deepcopy(self.doc)                                           # for duplicate, save&next
      if self.groupEdit:                                                                        # group update
        if 'name' in self.doc:
          del self.doc['name']
        if not self.doc['tags']:
          del self.doc['tags']                                # remove empty tags to not overwrite already set
        self.doc = {k:v for k,v in self.doc.items() if v}                             # filter out empty items
        for docID in self.allDocIDsCopy:
          self.comm.uiRequestTask.emit(Task.EDIT_DOC, {'doc':self.doc|{'id':docID}, 'newProjID':newProjID})
      elif 'id' in self.doc:                                                          # default update on item
        self.comm.uiRequestTask.emit(Task.EDIT_DOC, {'doc':self.doc, 'newProjID':newProjID})
      else:                                                                               # create new dataset
        self.comm.uiRequestTask.emit(Task.ADD_DOC, {'hierStack':newProjID, 'docType':self.doc['type'][0],
                                                    'doc':self.doc})
      self.doc = docBackup
      #!!! NO updates / redraw here since one does not know from where form came
      # e.g. sequential edit cannot have redraw here
      if command[0] in [Command.FORM_SAVE_NEXT, Command.FORM_SAVE_DUPL]:
        for delKey in [i for i in self.doc.keys() if i in ['id'] or i.startswith('meta')]: # delete these keys
          del self.doc[delKey]
        if command[0] is Command.FORM_SAVE_NEXT:
          self.comm.changeTable.emit(self.doc['type'][0], '')
      else:
        self.accept()                                                                                   #close
        self.close()

    elif command[0] is Command.FORM_ADD_KV and self.keyValueListL is not None:
      self.keyValueLabel.show()
      self.keyValueListW.show()
      self.keyLabels.append(QLineEdit(''))
      self.keyLabels[-1].setPlaceholderText('key')
      self.keyLabels[-1].setToolTip('Key (leave empty to delete key-value pair)')
      self.keyLabels[-1].setValidator(QRegularExpressionValidator('[a-zA-Z0-9]\\S+'))
      self.values.append(QLineEdit(''))
      self.values[-1].setPlaceholderText('value')
      self.keyValueListL.addRow(self.keyLabels[-1], self.values[-1])

    elif command[0] is Command.FORM_SHOW_DOC:
      doc = copy.deepcopy(self.doc)
      if 'image' in doc:
        del doc['image']
      messageWindow = ScrollMessageBox('Details', doc, style='QScrollArea{min-width:600 px; min-height:400px}')
      messageWindow.exec()

    else:
      logging.error('Unknown Command %s', command, exc_info=True)
    return


  def textChanged(self) -> None:
    """
    Text changed in editor -> update the display on the right
    """
    key = self.sender().accessibleName()                                                        # type: ignore
    getattr(self, f'textShow_{key}').setMarkdown(markdownEqualizer(getattr(self, f'textEdit_{key}').toPlainText()))
    return

  def delTag(self, _:str, tag:str) -> None:
    """
    Clicked button to delete tag
    """
    self.doc['tags'].remove(tag)
    self.updateTagsBar()
    return

  def addTag(self, tag:Union[str,int]) -> None:
    """
    Clicked to add tag. Since one needs to use indexChanged to allow the user to enter text, that delivers a int. To allow to differentiate
    between both comboboxes, they cannot be the same (both int), hence grades has to be textChanged

    Args:
      tag (str, int): index (otherTags) or text (grades)
    """
    if isinstance(tag, str):                                                                # text from grades
      if tag!='':
        self.doc['tags'] = [i for i in self.doc['tags'] if i[0]!='_']
        self.doc['tags'] += [f'_{len(tag)}']
        self.gradeChoices.setCurrentText('')
    elif tag<1:                                                                   # zero index from other-tags
      return
    else:
      tag = self.otherChoices.currentText()
      if tag not in self.doc['tags']:
        self.doc['tags'] += [tag]
      self.otherChoices.setCurrentText('')
    self.updateTagsBar()
    return

  def updateTagsBar(self) -> None:
    """
    After creation, tag removal, tag addition: update the information on screen
    """
    if self.tagsBarSubL is None:
      return
    #update tags
    for i in reversed(range(self.tagsBarSubL.count())):
      item = self.tagsBarSubL.itemAt(i)
      if item is not None: item.widget().setParent(None)
    for tag in (self.doc['tags'] if 'tags' in self.doc else []):
      if not re.match(r'^_\d$', tag):
        Label(tag, 'h3', self.tagsBarSubL, self.delTag, tag, 'click to remove')
    self.tagsBarSubL.addWidget(QWidget(), stretch=2)                                   #type: ignore[call-arg]
    #update choices in combobox
    tagsSet = {i for i in self.tagsAllList if i[0]!='_'}
    newChoicesList = ['']+list(tagsSet.difference([i for i in self.doc['tags'] if i[0]!='_']))
    self.otherChoices.clear()
    self.otherChoices.addItems(newChoicesList)
    return


  def changeTabs(self, idx:int) -> None:
    """
    Clicked on tabs to change to different one

    Args:
      idx (int): index of tab that was clicked
    """
    if idx == 0:
      self.btnAddKWPairs.show()
    else:
      self.btnAddKWPairs.hide()
    return

  def reject(self) -> None:
    """ Reject the dialog, stop the thread and disconnect signals """
    warnings.filterwarnings('ignore', category=RuntimeWarning)
    if hasattr(self.comm, 'backendThread') and self.comm.backendThread.worker is not None:
      self.comm.backendThread.worker.beSendDoc.disconnect(self.onGetData)
      self.comm.backendThread.worker.beSendTable.disconnect(self.onGetTable)
    self.checkThreadTimer.stop()
    super().reject()


  def accept(self) -> None:
    """ Accept the dialog, stop the thread and disconnect signals """
    warnings.filterwarnings('ignore', category=RuntimeWarning)
    if hasattr(self.comm, 'backendThread') and self.comm.backendThread.worker is not None:
      self.comm.backendThread.worker.beSendDoc.disconnect(self.onGetData)
      self.comm.backendThread.worker.beSendTable.disconnect(self.onGetTable)
    self.checkThreadTimer.stop()
    super().accept()


class Command(Enum):
  """ Commands used in this file """
  BUTTON_BAR       = 1
  CHANGE_EXTRACTOR = 2
  SAVE_IMAGE       = 3
  OPEN_FILEBROWSER = 4
  OPEN_EXTERNAL    = 5
  FOCUS_AREA       = 6
  AUTO_COMMENT     = 7
  FORM_SAVE        = 8
  FORM_CANCEL      = 9
  FORM_ADD_KV      = 10
  FORM_SHOW_DOC    = 11
  FORM_SAVE_NEXT   = 12
  FORM_SAVE_DUPL   = 13



""" Analyser add-on with much functionality: shows how complex add-ons can be. But do not have to."""
import itertools
import matplotlib
import matplotlib.pyplot
import numpy as np
import pandas as pd
from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar
from PySide6.QtCore import QSize
from PySide6.QtWidgets import (QComboBox, QDialog, QDialogButtonBox, QLineEdit,  # pylint: disable=no-name-in-module
                               QVBoxLayout)
from scipy import stats
from sklearn.metrics import r2_score
from pasta_eln.miscTools import MplCanvas, dfConvertColumns, isFloat
from pasta_eln.UI.guiStyle import Label, space, widgetAndLayout

# The following two variables are mandatory
description  = 'Default metadata plot'  #short description that is shown in the menu
reqParameter = {} #possibility for required parameters: like API-key, etc. {'API': 'text'}


class DataAnalyse(QDialog):
  """ Editor to change metadata of binary file """
  def __init__(self, df, widget):
    """
    Initialization

    Args:
      comm (Communicate): communication channel
      df (pandas dataframe): all data
    """
    super().__init__()
    prop_cycle = matplotlib.pyplot.rcParams['axes.prop_cycle']
    self.colors = itertools.cycle(prop_cycle.by_key()['color'])
    self.setStyleSheet(f"QLineEdit, QComboBox {{  {widget.comm.palette.get('secondaryText','color')} }}")
    self.df = df
    columns = []
    for col in df.columns:
      if pd.api.types.is_numeric_dtype(df[col].dtype):
        columns.append(col)

    # GUI elements
    self.setWindowTitle('Analyze metadata')
    self.setMinimumWidth(1050)  #set size to match 4 blocks of 8bytes
    mainL = QVBoxLayout(self)
    mainL.setSpacing(space['s'])

    #graph
    self.graphW, graphL = widgetAndLayout('V', None)
    self.graph = MplCanvas(self, width=5, height=4, dpi=100)
    self.graphToolbar = NavigationToolbar(self.graph, self)
    self.graphToolbar.setIconSize(QSize(24, 24))
    self.graphToolbar.setStyleSheet('QToolButton {min-width: 18px; min-height: 24px;}')
    graphL.addWidget(self.graphToolbar)
    graphL.addWidget(self.graph)
    mainL.addWidget(self.graphW, stretch=1)

    #Graph type
    _, rowType = widgetAndLayout('H', mainL, 'm', 's', '0', 's')
    Label('graph type:','h2',rowType)
    self.typeCB = QComboBox()
    self.typeCB.addItems(['x-y plot','histogram'])
    self.typeCB.currentTextChanged.connect(self.changePlotType)
    rowType.addWidget(self.typeCB)
    rowType.addSpacing(1)
    self.subtypeCB = QComboBox()
    self.subtypeCB.addItems(['points','points & lines','lines'])
    self.subtypeCB.currentTextChanged.connect(self.refresh)
    rowType.addWidget(self.subtypeCB)
    rowType.addSpacing(1)

    #X-Axis
    self.rowXW, rowXL = widgetAndLayout('H', mainL, 'm', 's', '0', 's')
    Label('x-axis:','h2',rowXL)
    self.xAxisCB = QComboBox()
    self.xAxisCB.addItems(columns)
    self.xAxisCB.currentTextChanged.connect(self.refresh)
    rowXL.addWidget(self.xAxisCB)
    Label('min:','',rowXL)
    self.xAxisMin = QLineEdit('')
    self.xAxisMin.textChanged.connect(self.refresh)
    rowXL.addWidget(self.xAxisMin)
    Label('max:','',rowXL)
    self.xAxisMax = QLineEdit('')
    self.xAxisMax.textChanged.connect(self.refresh)
    rowXL.addWidget(self.xAxisMax)
    Label('label:','',rowXL)
    self.xAxisLabel = QLineEdit('')
    self.xAxisLabel.textChanged.connect(self.refresh)
    rowXL.addWidget(self.xAxisLabel)

    #Y-Axis
    self.rowYW, rowYL = widgetAndLayout('H', mainL, 'm', 's', '0', 's')
    Label('y-axis:','h2',rowYL)
    self.yAxisCB = QComboBox()
    self.yAxisCB.addItems(columns)
    self.yAxisCB.currentTextChanged.connect(self.refresh)
    rowYL.addWidget(self.yAxisCB)
    Label('min:','',rowYL)
    self.yAxisMin = QLineEdit('')
    self.yAxisMin.textChanged.connect(self.refresh)
    rowYL.addWidget(self.yAxisMin)
    Label('max:','',rowYL)
    self.yAxisMax = QLineEdit('')
    self.yAxisMax.textChanged.connect(self.refresh)
    rowYL.addWidget(self.yAxisMax)
    Label('label:','',rowYL)
    self.yAxisLabel = QLineEdit('')
    self.yAxisLabel.textChanged.connect(self.refresh)
    rowYL.addWidget(self.yAxisLabel)

    #Coloring
    _, rowCL = widgetAndLayout('H', mainL, 'm', 's', '0', 's')
    Label('color:','h2',rowCL)
    self.cAxisCB = QComboBox()
    self.cAxisCB.addItems(df.columns)
    self.cAxisCB.currentTextChanged.connect(self.refresh)
    rowCL.addWidget(self.cAxisCB)
    Label('min:','',rowCL)
    self.cAxisMin = QLineEdit('')
    self.cAxisMin.textChanged.connect(self.refresh)
    rowCL.addWidget(self.cAxisMin)
    Label('max:','',rowCL)
    self.cAxisMax = QLineEdit('')
    self.cAxisMax.textChanged.connect(self.refresh)
    rowCL.addWidget(self.cAxisMax)
    Label('label:','',rowCL)
    self.cAxisLabel = QLineEdit('')
    self.cAxisLabel.textChanged.connect(self.refresh)
    rowCL.addWidget(self.cAxisLabel)

    #final button box
    buttonBox = QDialogButtonBox(QDialogButtonBox.Cancel)
    buttonBox.clicked.connect(self.cancel)
    mainL.addWidget(buttonBox)
    self.refresh()

  def changePlotType(self):
    """ change the main plot-type: changes also the list in subType and hence refreshes the graph"""
    if self.typeCB.currentText()=='x-y plot':
      self.rowYW.show()
      self.subtypeCB.clear()
      self.subtypeCB.addItems(['points','points & lines','lines'])
    if self.typeCB.currentText()=='histogram':
      self.rowYW.hide()
      self.subtypeCB.clear()
      self.subtypeCB.addItems(['histogram','cumulative distribution','distribution'])


  def refresh(self):
    """ repaint graph """
    try:
      self.df.sort_values(self.xAxisCB.currentText(), inplace=True)
    except KeyError:
      pass
    x = self.df[self.xAxisCB.currentText()]
    y = self.df[self.yAxisCB.currentText()]
    c = self.df[self.cAxisCB.currentText()]
    try:
      fit = np.polyfit(x, y, 1)
    except np.linalg.LinAlgError:
      fit = [np.nan, np.nan]
    xMin = None if self.xAxisMin.text()=='' or not isFloat(self.xAxisMin.text()) else float(self.xAxisMin.text())
    xMax = None if self.xAxisMax.text()=='' or not isFloat(self.xAxisMax.text()) else float(self.xAxisMax.text())
    yMin = None if self.yAxisMin.text()=='' or not isFloat(self.yAxisMin.text()) else float(self.yAxisMin.text())
    yMax = None if self.yAxisMax.text()=='' or not isFloat(self.yAxisMax.text()) else float(self.yAxisMax.text())
    infoText = ''

    matplotlib.pyplot.close('all')
    self.graph.axes.cla()
    #SCATTER PLOT
    if self.typeCB.currentText()=='x-y plot':
      # fit
      self.graph.axes.plot(x, np.polyval(fit, x), 'k--')
      text = f'y={fit[0]:.3f}*x+{fit[1]:.3f}' # : $R^2$={round(r2_score(y, np.polyval(fit, x)), 2)}' #Comment in and install sklearn
      self.graph.axes.text(x.mean(), np.polyval(fit, x.mean()), text, verticalalignment='top')
      self.graph.axes.set_ylim(bottom=yMin, top=yMax)
      self.graph.axes.set_ylabel(self.yAxisLabel.text() or self.yAxisCB.currentText())
      # real plot
      subType = self.subtypeCB.currentText()
      plotType = 'o' if subType=='points' else 'o-' if subType=='points & lines' else '-'
      for ci in np.unique(c):
        mask = c==ci
        self.graph.axes.plot(x[mask],y[mask], plotType, label=ci)

    # HISTOGRAM
    if self.typeCB.currentText()=='histogram':
      for ci in np.unique(c):
        mask = c==ci
        label = f'{ci} {np.mean(x[mask]):.2e}$\pm${np.std(x[mask]):.2e}'
        if self.subtypeCB.currentText()=='histogram':
          self.graph.axes.hist(x[mask], 20, label=label, alpha=0.5)
          self.graph.axes.set_ylabel('count')
        if self.subtypeCB.currentText()=='cumulative distribution':
          self.graph.axes.ecdf(x[mask], label=label)
          self.graph.axes.set_ylabel('cumulative distribution')
        if self.subtypeCB.currentText()=='distribution':
          res = stats.ecdf(x[mask])
          y_ = (res.cdf.probabilities[1:]-res.cdf.probabilities[:-1])/(res.cdf.quantiles[1:]-res.cdf.quantiles[:-1])
          x_ = (res.cdf.quantiles[1:]+res.cdf.quantiles[:-1])/2
          w_ = (res.cdf.quantiles[1:]-res.cdf.quantiles[:-1])
          color = next(self.colors)
          self.graph.axes.bar(x_, y_, width=w_, label=label, facecolor='none', linestyle='-', linewidth=1, edgecolor=color)
          self.graph.axes.set_ylabel('relative distribution')
      ksStats = stats.ks_2samp(x[mask], x[~mask])
      infoText = f'p-value={ksStats.pvalue:.3f}'

    # for all plots
    self.graph.axes.legend(title = self.cAxisLabel.text() or self.cAxisCB.currentText())
    xlim = self.graph.axes.set_xlim(left=xMin, right=xMax)
    ylim = self.graph.axes.get_ylim()
    self.graph.axes.set_xlabel(self.xAxisLabel.text() or self.xAxisCB.currentText())
    if infoText:
      self.graph.axes.text((xlim[0]+xlim[1])/2, ylim[1], infoText , horizontalalignment='center', verticalalignment='top')
    self.graph.axes.tick_params(axis='both', direction='in')
    self.graph.figure.tight_layout()
    self.graph.draw()
    self.graphToolbar.show()
    self.graph.show()
    return

  def cancel(self, btn):
    """ cancel selectedList to configuration and exit """
    self.reject()
    return


def main(comm, df, widget, parameter={}):
    """ main function: has to exist and is called by the menu
    Args:
        comm (communication): pasta communication layer
        df (Dataframe): dataframe with the data to plot
        widget (QWidget): allows to create new gui dialogs
        parameter (dict): ability to pass parameters

    Returns:
        bool: success
    """
    dataAnalyse = DataAnalyse(dfConvertColumns(df, 10), widget)
    dataAnalyse.exec()



"""Input and output functions towards the .eln file-format"""
import copy
import hashlib
import json
import logging
import re
import shutil
import sys
import time
import uuid
from datetime import datetime
from pathlib import Path
from typing import Any
from zipfile import ZIP_DEFLATED, ZipFile, ZipInfo
import requests
from anytree import Node
from pasta_eln import __version__, minisign
from ..fixedStringsJson import CONF_FILE_NAME
from ..miscTools import flatten
from ..textTools.html2markdown import html2markdown
from ..textTools.stringChanges import camelCase
from .backend import Backend
from .htmlString import htmlEnd, htmlStart, importantKeys

# .eln file: common between all ELNs
# - can be exported / imported generally; not a 1:1 backup (just zip it)
# - should be able to recreate the exported -> imported data (using the common addDoc)
# - externalID is the only content that cannot be recreated
#   - not sure how important this information is: different groups have different externalIDs
#   - if every it becomes important: just add json (that maps identifier to externalID) as additional file

# Idea: ELN from other vendor -> import -> export: should be the same
# - not because some information is added (pasta's identifier)
# - some information is changed (what I understand as measurement is not the same as elab's experiment which is more a folder in pasta
# - some information is deleted (data privacy related)

# GENERAL TERMS IN ro-crate-metadata.json (None: this entry is not saved and will be recreated upon import)
pasta2json:dict[str,Any] = {
  'id'          : 'identifier',
  'branch'      : None,
  'client'      : None,
  'dateModified': 'dateModified',
  'dateCreated' : 'dateCreated',
  'name'        : 'name',
  'tags'        : 'keywords',
  'type'        : 'genre',
  'image'       : None,
  'comment'     : 'description',
  'content'     : 'text',
  '.links'      : 'mentions',
  'shasum'      : None,
  'user'        : None,
  'externalId'  : None,
  'gui'         : None,
  'dateSync'    : None
}
json2pasta:dict[str,Any] = {v:k for k,v in pasta2json.items() if v is not None}

METADATA_FILE = 'ro-crate-metadata.json'


##########################################
###               IMPORT               ###
##########################################

def importELN(backend:Backend, elnFileName:str, projID:str) -> tuple[str,dict[str,Any]]:
  '''
  import .eln file from other ELN or from PASTA

  Args:
    backend (backend): backend
    elnFileName (str): name of file
    projID (str): project to import data into

  Returns:
    str: success message, statistics
  '''
  elnName = ''
  statistics:dict[str,Any] = {}
  with ZipFile(elnFileName, 'r', compression=ZIP_DEFLATED) as elnFile:
    files = elnFile.namelist()
    baseFolderSet = {Path(i).parts[0] for i in files}
    if len(baseFolderSet) != 1:
      logging.error('eln file has multiple top-level directories: %s. Cannot process',str(baseFolderSet))
      return f'ERROR: eln file has multiple top-level directories: {baseFolderSet}. Cannot process',{}
    dirName=Path(files[0]).parts[0]
    statistics['num. files'] = len([i for i in files if Path(i).parent!=Path(dirName)])
    if f'{dirName}/ro-crate-metadata.json' not in files:
      logging.error('ro-crate does not exist in folder. EXIT')
      return 'ERROR: ro-crate does not exist in folder. EXIT',{}
    graph = json.loads(elnFile.read(f'{dirName}/ro-crate-metadata.json'))['@graph']
    listAllTypes = [i['@type'] for i in graph if isinstance(i['@type'],str)]
    statistics['types'] = {i:listAllTypes.count(i) for i in listAllTypes}

    #find information from master node
    rocrateNode = [i for i in graph if i['@id'].endswith('ro-crate-metadata.json')][0]
    if 'sdPublisher' in rocrateNode:
      publisherNode = rocrateNode['sdPublisher']
      if 'name' not in publisherNode:
        publisherNode = [i for i in graph if i['@id']==rocrateNode['sdPublisher']['@id']][0]
      elnName = publisherNode['name']
    logging.info('Import %s', elnName)
    if not projID:
      return 'FAILURE: YOU CANNOT IMPORT AS PROJECT IF NON PASTA-ELN FILE',{}
    backend.changeHierarchy(projID)
    childrenStack = [0]
    mainNode    = [i for i in graph if i['@id']=='./'][0]
    # clean subchildren from mainNode: see https://github.com/TheELNConsortium/TheELNFileFormat/issues/98
    parentNodes = {i['@id'] for i in mainNode['hasPart']}
    for nodeAny in graph:
      if nodeAny['@id']=='./':
        continue
      children = {i['@id'] for i in nodeAny.get('hasPart',[])}
      parentNodes = parentNodes.difference(children)
    mainNode['hasPart'] = [{'@id':i} for i in parentNodes]

    ################
    # subfunctions #
    ################
    def json2pastaFunction(inputData:dict[str,Any]) -> tuple[dict[str,Any], str, list[Any], str]:
      """
      convert json data to pastaStyle
        - clean all the @id, @type

      Args:
        inputData (dict): input data in document

      Returns:
        dict: output data as document
        str: id in eln=path in zip file
        list: list of children
        str: data or dataset
      """
      doc = {}
      elnID = inputData['@id'][2:] if inputData['@id'][:2]=='./' else inputData['@id']
      children = inputData.pop('hasPart') if 'hasPart' in inputData else []
      dataType = inputData['@type']
      encodingFormat = ''
      for key, value in inputData.items():
        if key in ['@id','@type','hasPart','author','contentSize', 'sha256']:
          continue
        if key in json2pasta:                                                           #use known translation
          doc[json2pasta[key]] = value
        elif key == 'encodingFormat':
          encodingFormat = inputData[key]
        elif value:                                                 #keep name, only prevent causes for errors
          doc[f'.{key}'] = value
      # change into Pasta's native format
      if isinstance(doc.get('comment',''), (dict,list)):
        if isinstance(doc['comment'], dict):
          doc['comment'] = [doc['comment']]
        doc['comment'] = '\n\n'.join([i['text'] for i in doc['comment']])
      if '.comment' in doc:
        doc['comment'] = doc.get('comment','') + \
                         '\n\n'.join(['']+[f'{i["dateCreated"]}:\n{i["text"]}' for i in doc.pop('.comment')])
      if encodingFormat=='text/html':
        if 'comment' in doc:
          doc['comment'] = html2markdown(doc['comment'])
        if 'content' in doc:
          doc['content'] = html2markdown(doc['content'])
      if 'tags' in doc:
        doc['tags'] = [i.strip() for i in doc['tags'].split(',')]
      else:
        doc['tags'] = []
      if 'id' in doc:
        doc['.oldIdentifier'] = doc.pop('id')
      doc['.elnIdentifier'] = elnID
      if (children and ('type' not in doc or doc['type']!='x1')):
        doc['type'] = 'x1'
      if 'type' not in doc:
        doc['type'] = ''
      if doc['type']=='folder':
        doc['type'] = 'x1'
      # change .variable measured into pastaSystem
      variableMeasured = doc.pop('.variableMeasured',[])
      if variableMeasured is not None:
        for data in variableMeasured:
          propertyID = data['propertyID'] if '.' in data['propertyID'] and ' ' not in data['propertyID'] else f"imported.{camelCase(data['propertyID'])}"
          # print('propertyID', propertyID, elnName)
          doc[propertyID] = (data['value'], data.get('unitText',''), data.get('description',''), data.get('mentions',''))
      # final cleaning
      for key, value in doc.items():
        if isinstance(value, list) and len(value)>0 and isinstance(value[0], dict):
          newValue = []
          for item in copy.deepcopy(value):
            del item['@id']
            del item['@type']
            newValue.append(item)
          if len(newValue)==1:
            newValue = newValue[0]
          doc[key] = newValue
        elif isinstance(value, dict):
          for keyI in list(value.keys()):
            if keyI.startswith('@'):
              newKey = keyI[1:]
              newKey = 'imported_'+newKey if newKey in ['id','type','branch'] else newKey
              value[newKey] = value.pop(keyI)
      # print(f'Node becomes: {json.dumps(doc, indent=2)}')
      return doc, elnID, children, dataType


    def processPart(part:dict[str,str]) -> int:
      """
      recursive function call to process this node

      Args:
        part (dict): dictionary containing this data content

      Returns:
        int: number of documents added
      """
      addedDocs = 1
      if not isinstance(part, dict):              #leave these tests in since other .elns might do funky stuff
        logging.error('in part %s', part, exc_info=True)
        return 0
      # print('\nProcess: '+part['@id'])
      # find next node to process
      docS = [i for i in graph if '@id' in i and i['@id']==part['@id']]
      if len(docS)!=1 or backend.cwd is None:
        logging.error('zero or multiple nodes with same id %s or cwd is None in %s', docS, part['@id'], exc_info=True)
        return -1
      # pull all subentries (variableMeasured, comments, ...) into this dict: do not pull hasPart-entries in
      keys = [k for k,v in docS[0].items() if k!='hasPart' and (isinstance(v,dict) or (isinstance(v,list) and len(v)>0 and isinstance(v[0], dict)))]
      for key in keys:
        value = docS[0].get(key, [])
        if isinstance(value, list):
          items = [i['@id'] for i in value]
        else:
          items = [value['@id']]
        # print(f'Replace {key} -entries using ids: {items}')
        try:
          docS[0][key] = [ [j for j in graph if j['@id']==i][0] for i in items]
        except Exception:
          # Causes for exception
          # - @id is a link to external IRI (http...). Hence it is not in the file
          # - the graph is not fully flattened: @id is in a leaf of a node but not a separate note
          docS[0][key] = value
          logging.warning('Could not replace %s-entries using ids: %s', key, items)
      # convert to Pasta's style
      doc, elnID, children, dataType = json2pastaFunction(docS[0])
      if elnName == 'PASTA ELN' and elnID.startswith('http') and ':/' in elnID:
        fullPath = None
      else:
        fullPath = backend.basePath/backend.cwd/elnID.split('/')[-1]
      # Copy file onto hard disk
      if fullPath is not None and f'{dirName}/{elnID}' in elnFile.namelist() and \
          not [i for i in elnFile.infolist() if f'{dirName}/{elnID}'==i.filename][0].is_dir():#prevent folders
        target = open(fullPath, 'wb')
        source = elnFile.open(f'{dirName}/{elnID}')
        with source, target:                                          #extract one file to its target directly
          shutil.copyfileobj(source, target)
      # FOR ALL ELNs
      if elnName == 'PASTA ELN':
        docType = doc['type']
      else:
        if dataType.lower()=='dataset':
          docType = 'x0' if not projID and len(elnID.split('/'))==0 else 'x1'
        else:
          docType = ''
      if docType=='x0':
        backend.hierStack = []
      childrenStack[-1] += 1
      doc['childNum'] = childrenStack[-1]
      qrCodes = [v[0] for k,v in doc.items() if 'qrCodes.' in k]
      doc = {k:v for k,v in doc.items() if 'qrCodes.' not in k}
      doc['qrCodes'] = qrCodes
      # print(f'Want to add doc:{doc} with type:{docType} and cwd:{backend.cwd}')
      try:
        docID = backend.addData(docType, doc)['id']
      except Exception:
        logging.error('Cannot process %s', json.dumps(doc,indent=2), exc_info=True)
        docID = None
      if docID[0]=='x':
        backend.changeHierarchy(docID)
        childrenStack.append(0)
        with open(backend.basePath/backend.cwd/'.id_pastaELN.json','w', encoding='utf-8') as f:#local path, update in any case
          f.write(json.dumps(backend.db.getDoc(docID)))
        # children, aka recursive part
        for child in children:
          if child['@id'].endswith('/metadata.json') or child['@id'].endswith('_metadata.json'):#skip own metadata
            continue
          try:
            addedDocs += processPart(child)
          except Exception:
            logging.error('Cannot process child %s', json.dumps(child,indent=2), exc_info=True)
        backend.changeHierarchy(None)
        childrenStack.pop()
      return addedDocs

    ######################
    #main function
    #iteratively go through list
    addedDocuments = 0
    for part in mainNode['hasPart']:
      try:
        addedDocuments += processPart(part)
      except Exception:
        logging.error('Cannot process main part %s', json.dumps(part,indent=2), exc_info=True)
  #return to home stack and path
  backend.cwd = Path(backend.basePath)
  backend.hierStack = []
  return f'Success: imported {str(addedDocuments)} documents from file {elnFileName} from ELN {elnName}', statistics




##########################################
###               EXPORT               ###
##########################################

def exportELN(backend:Backend, projectIDs:list[str], fileName:str, dTypes:list[str], verbose:bool=False) -> str:
  """
  export eln to file

  Args:
    backend (backend): PASTA backend instance
    projectIDs (list): list of docIds of projects
    fileName (str): fileName which to use for saving
    dTypes (list): list of strings which should be included in the output, alongside folders x0 & x1
    verbose (bool): verbose

  Returns:
    str: report of exportation
  """
  # Initialize variables
  fileName = fileName if fileName.endswith('.eln') else f'{fileName}.eln'
  logging.info('Create eln file %s',fileName)
  dirNameGlobal = fileName.split('/')[-1][:-4]
  with ZipFile(fileName, 'w', compression=ZIP_DEFLATED) as elnFile:
    graph: list[dict[str,Any]] = []

    def mkDirectory(path:str) -> None:
      """ create directory in zip file

      Args:
        path (str): path to create
      """
      if path.startswith('./'):
        fullPath = f'{dirNameGlobal}/{path[2:]}'
      else:
        fullPath = f'{dirNameGlobal}/{path}'
      if sys.version_info >= (3, 11):
        elnFile.mkdir(fullPath)
      else:
        dirInfo = ZipInfo(fullPath if fullPath.endswith('/') else f'{fullPath}/')
        dirInfo.date_time = time.localtime(time.time())[:6]
        dirInfo.external_attr = 0o40775 << 16                                                     # drwxrwxr-x
        elnFile.writestr(dirInfo, '')

    def processNode(node:Node) -> str:
      """
      Recursive function to translate the hierarchical node into a tree-node

      Args:
        node (Anytree.Node): anytree node with properties
            name (incl. that of parents), childNum, docType, gui, id

      Returns:
        str: tree node id
      """
      # create node properties
      docDB = backend.db.getDoc(node.id)
      docELN:dict[str,Any] = {'encodingFormat': 'text/markdown'}
      docSupp = {}
      for key, value in docDB.items():
        if key in pasta2json:
          if pasta2json[key] is not None:
            docELN[pasta2json[key]] = value
        else:
          docSupp[key] = value
      # clean individual entries of docELN
      if 'keywords' in docELN:
        docELN['keywords'] = ','.join(docELN['keywords'])
      if docELN['genre'][0] == '-':
        del docELN['genre']
      elif docELN['genre'][0][0]=='x':
        docELN['genre'] = 'folder'
      else:
        docELN['genre'] = '/'.join(docELN['genre'])
      docELN = {k:v for k,v in docELN.items() if v}
      # include children, hasPart
      hasPart = []
      if (node.docType[0] not in dTypes) and node.docType[0][0]!='x':
        return ''
      for child in node.children:
        res = processNode(child)
        if res:
          hasPart.append(res)
      if node.docType[0][0]=='x' and hasPart:
        docELN['hasPart'] = [{'@id':i} for i in hasPart]
      # include path and @id
      if docDB['type'][0] == 'x0':
        branch   = docDB['branch'][0]
      else:
        branch   = [i for i in docDB['branch'] if node.parent.id in i['stack']][0]
      path =  f"{branch['path']}/" if docDB['type'][0][0]=='x' else branch['path']
      if path is None:
        path = f"{dirNameProject}/{docDB['id']}"
      if path.startswith('http'):
        docELN['@id'] = path
        docELN['url'] = path
      else:
        docELN['@id'] = f'./{path}'
      # include content size, etc
      fullPath = backend.basePath/path
      if path is not None and fullPath.exists() and fullPath.is_file():
        with open(fullPath, 'rb') as fIn:
          fileContent = fIn.read()
          docELN['contentSize'] = str(len(fileContent))
          docELN['sha256']      = hashlib.sha256(fileContent).hexdigest()
        # copy data-files
        elnFile.write(str(fullPath), f'{dirNameGlobal}/{path}')
        docELN['@type'] = 'File'
      elif path is not None and fullPath.exists() and fullPath.is_dir():
        mkDirectory(docELN['@id'][:-1])
        docELN['@type'] = 'Dataset'
      elif path.startswith('http'):
        response = requests.get(path, timeout=10)
        if response.ok:
          docELN['contentSize'] = str(response.headers['content-length'])
          docELN['sha256']      = hashlib.sha256(response.content).hexdigest()
        else:
          print(f"Info: could not get file {path}")
        docELN['@type'] = 'File'
      elif '@type' not in docELN:                                                        #samples will be here
        docELN['@type'] = 'Dataset'
        docELN['@id'] = docELN['@id'] if docELN['@id'].endswith('/') else f"{docELN['@id']}/"
        mkDirectory(docELN['@id'][:-1])
      # move docSupp into separate nodes
      variableMeasuredIDs = []
      for kObject, v in flatten(docSupp).items():
        name = ' \u2192 '.join([i.replace('_','').replace('-','').capitalize() for i in str(kObject).split('.')])
        varMeasured = {'value':v, 'propertyID':kObject, 'name':name, '@type': 'PropertyValue','@id':f'{docELN["@id"]}_{kObject}'}
        if isinstance(v, tuple):
          varMeasured['value'] = v[0]
          if v[1]:
            varMeasured['unitText'] = v[1]
        if isinstance(v, tuple):
          varMeasured['value'] = v[0]
          if v[1]:
            varMeasured['unitText'] = v[1]
          if v[2]:
            varMeasured['description'] = v[2]
          if v[3]:
            varMeasured['identifier'] = v[3]
        graph.append(varMeasured)
        variableMeasuredIDs.append({'@id':f'{docELN["@id"]}_{kObject}'})
      if variableMeasuredIDs:
        docELN['variableMeasured'] = variableMeasuredIDs
      graph.append(docELN)
      return docELN['@id']


    def createNodeHTML(node:dict[str,Any]) -> str:
      """
      create HTML representation of a node
      Args:
        node (dict): node to represent

      Returns:
        str: HTML representation
      """
      output = '<h3 id="{i}">{n} ({t})</h3>\n'.format(i=node['@id'],n=node.get('name',node['@id']),# pylint: disable=consider-using-f-string
                                                      t=node['@type'])
      output += '<table>\n<colgroup><col style="width: 300px;"><col></colgroup>\n<tbody>'
      for key in importantKeys:
        if key not in node:
          continue
        value = node[key]
        if key == 'hasPart':
          value = ', <br>'.join(['<a href="#{i}">{n}</a>'.format(i=node['@id'],n=node.get('name',node['@id']))# pylint: disable=consider-using-f-string
                                 for i in value])
        if key == 'identifier':
          if isinstance(value, str) and value.startswith('http'):
            value = f'<a href="{value}">{value}</a>'
          else:
            continue
        output += f'<tr><td>{key}</td><td>{value}</td></tr>\n'
      output += '</tbody></table>\n'
      output += '<details><summary>Details</summary>\n'
      output += '<table>\n<colgroup><col style="width: 300px;"><col></colgroup>\n<tbody>'
      for key, value in node.items():
        if key in importantKeys and key != 'identifier':
          continue
        output += f'<tr><td>{key}</td><td>{value}</td></tr>\n'
      output += '</tbody></table>\n'
      output += '</details>'
      output += '<hr>\n'
      return output


    def createHTML(graph:list[dict[str,Any]]) -> str:
      """ create HTML representation of the graph
      Args:
        graph (list): list of nodes

      Returns:
        str: HTML representation
      """
      output = createNodeHTML([i for i in graph if i['@id']=='./'][0])
      for node in graph:
        if node['@id']=='./' or node['@id'].endswith('ro-crate-metadata.json'):
          continue
        output += createNodeHTML(node)
      return htmlStart + output + htmlEnd.replace('___VERSION___', __version__).replace('___DATE___', str(datetime.now()))


    # for each project, append to graph
    masterParts = []
    for projectID in projectIDs:
      docProject = backend.db.getDoc(projectID)
      dirNameProject = docProject['branch'][0]['path']
      listHier, _ = backend.db.getHierarchy(projectID, allItems=False)#error not handled since should not occur during export
      processNode(listHier)
      masterParts.append(f'./{dirNameProject}/')

    # all items have to appear in hasPart of ./ -> masterParts are changed
    # https://github.com/TheELNConsortium/TheELNFileFormat/issues/98
    somethingChanged = True                                                                #starting condition
    nodesProcessed = set()
    while somethingChanged:
      somethingChanged = False
      for node in copy.deepcopy(masterParts):
        if node in nodesProcessed:
          continue
        nodesProcessed.add(node)
        possNodes = [i for i in graph if i['@id']==node]
        if len(possNodes)==1:
          idx = masterParts.index(node)
          variables:list[str] = []
          # variables = [i['@id'] for i in possNodes[0].get('variableMeasured',[])] #variables go not into ./
          children            = [i['@id'] for i in possNodes[0].get('hasPart',[])]
          masterParts = masterParts[:idx+1] + variables + children + masterParts[idx+1:]
          if variables or children:
            somethingChanged = True

    # FOR ALL PROJECTS
    # ------------------- create ro-crate-metadata.json header -----------------------
    index:dict[str,Any] = {}
    index['@context']= 'https://w3id.org/ro/crate/1.1/context'
    # master node ro-crate-metadata.json
    graphMaster:list[dict[str,Any]] = []
    graphMisc:list[dict[str,Any]] = []
    masterNodeInfo = {'@id': 'ro-crate-metadata.json', '@type': 'CreativeWork', 'about': {'@id': './'},
        'conformsTo': {'@id': 'https://w3id.org/ro/crate/1.1'}, 'version': '1.0',
        'additionalType': 'https://purl.archive.org/purl/elnconsortium/eln-spec/1.1',
        'datePublished':datetime.now().isoformat(), 'dateCreated': datetime.now().isoformat(),
        'sdPublisher': {'@id': 'PASTA-ELN'}}
    graphMaster.append(masterNodeInfo)
    masterNodeInfo2 = {'@id':'PASTA-ELN','@type': 'Organization', 'name': 'PASTA ELN',
            'logo': 'https://raw.githubusercontent.com/PASTA-ELN/desktop/main/pasta.png',
            'slogan': 'The favorite ELN for experimental scientists',
            'url': 'https://github.com/PASTA-ELN/', 'version': __version__
    }
    graphMaster.append(masterNodeInfo2)
    authorNodes = []
    for author in backend.configuration['authors']:
      affiliationNodes = []
      for affiliation in author['organizations']:
        affiliationId    = f"affiliation_{affiliation['organization']}"
        if affiliationId not in graphMaster:
          data = {'@id':affiliationId, '@type':'Organization', 'name':affiliation['organization']}
          if 'rorid' in affiliation and affiliation['rorid']:
            data['identifier'] = f"https://ror.org/{affiliation['rorid']}"
          graphMaster.append(data)
          affiliationNodes.append({'@id':affiliationId})
      authorID = f"author_{author['first']}_{author['last']}"
      graphMaster.append({'@id':authorID, '@type':'Person', 'givenName': author['first'], 'familyName': author['last'],
                          'honorificPrefix': author['title'], 'email': author['email'], 'identifier': f"https://orcid.org/{author['orcid']}",
                          'worksFor': affiliationNodes})
      authorNodes.append({'@id':authorID})
    masterNodeRoot:dict[str,Any] = {'@id': './', '@type': 'Dataset', 'hasPart': [{'@id':i} for i in masterParts],
        'name': 'Exported from PASTA ELN', 'description': 'Exported content from PASTA ELN',
        'license':'https://creativecommons.org/licenses/by-nc-sa/4.0/', 'datePublished':datetime.now().isoformat()}
    if authorNodes:
      masterNodeRoot = masterNodeRoot | {'creator': authorNodes}
    graphMaster.append(masterNodeRoot)

    #finalize file
    index['@graph'] = graphMaster+graph+graphMisc
    elnFile.writestr(f'{dirNameGlobal}/ro-crate-metadata.json', json.dumps(index, indent=2))
    elnFile.writestr(f'{dirNameGlobal}/ro-crate-preview.html',  createHTML(index['@graph']))
    # find nodes that could be defined
    possDefined = [(i['@id'], i['@type'],i.get('identifier','')) for i in index['@graph'] if not re.search(r'\w-[0-9a-f]{32}',i.get('identifier',''))]
    definedStr = '  -'+'\n  -'.join([str(i) for i in possDefined if i[2]!=''])
    logging.info('Defined identifiers:\n%s',definedStr)
    undefinedProperty = '  -'+'\n  -'.join([i[0] for i in possDefined if i[2]=='' and i[1]=='PropertyValue'])
    logging.info('Undefined PropertyValue identifiers:\n%s',undefinedProperty)
    undefinedNonProperty = '  -'+'\n  -'.join([i[0] for i in possDefined if i[2]=='' and i[1]!='PropertyValue' and i[0] not in ('./','ro-crate-metadata.json')])
    logging.info('Undefined NON-PropertyValue identifiers:\n%s',undefinedNonProperty)

    #sign file
    if 'signingKeyPair' not in backend.configuration or not backend.configuration['signingKeyPair']:
      #create a key-pair of secret and public key and save it locally
      keyPairRaw = minisign.KeyPair.generate()
      keyPair    = {'id':str(uuid.uuid4()), 'secret':bytes(keyPairRaw.secret_key).decode(), 'public':bytes(keyPairRaw.public_key).decode()}
      backend.configuration['signingKeyPair'] = keyPair
      with open(Path.home()/CONF_FILE_NAME, 'w', encoding='utf-8') as fConf:
        json.dump(backend.configuration, fConf, ensure_ascii=False, indent=2)
    keyPair = backend.configuration['signingKeyPair']
    secretKey = minisign.SecretKey.from_bytes(keyPair['secret'].encode())
    comment   = {'pubkey_url':f'https://raw.githubusercontent.com/PASTA-ELN/Signatures/main/{keyPair["id"]}.pub',
                 'name': f"{backend.configuration['authors'][0]['title']} {backend.configuration['authors'][0]['first']} {backend.configuration['authors'][0]['last']}",
                 'email': backend.configuration['authors'][0]['email'],
                 'orcid': backend.configuration['authors'][0]['orcid']}
    signature = secretKey.sign(json.dumps(index).encode(), trusted_comment=json.dumps(comment))
    elnFile.writestr(f'{dirNameGlobal}/ro-crate-metadata.json.minisig', bytes(signature).decode())
    elnFile.writestr(f'{dirNameGlobal}/ro-crate.pubkey', keyPair['public'])
  # end writing zip file

  # temporary json output
  if verbose:
    with open(f'{fileName[:-3]}json', 'w', encoding='utf-8') as fOut:
      fOut.write( json.dumps(index, indent=2) )
  return f'Success: exported {len(graph)} nodes into file {fileName}. '\
    'Count of defined nodes: {n1}, count of undefined properties: {n2}, count of undefined others: {n3}'.format(
    n1=len(definedStr.split('\n')),
    n2=len(undefinedProperty.split('\n')),
    n3=len(undefinedNonProperty.split('\n')))


def validateSignature(fileName:str) -> bool:
  """ Test if signature is valid

  Args:
    fileName (str): file name to test against

  Returns:
    bool: success / failure of test
  """
  print(f'Test signature of : {fileName}')
  with ZipFile(fileName, 'r', compression=ZIP_DEFLATED) as elnFile:
    metadataJsonFile = [i for i in elnFile.namelist() if i.endswith(METADATA_FILE)][0]
    with elnFile.open(metadataJsonFile) as fIn:
      metadataData = fIn.read()
    dirName          = metadataJsonFile.split('/')[0]
    signatureFile    = f'{dirName}/{METADATA_FILE}.minisig'
    with elnFile.open(signatureFile) as fIn:
      signature = minisign.Signature.from_bytes(fIn.read())
    publicKeyFile    = f'{dirName}/ro-crate.pubkey'
    with elnFile.open(publicKeyFile) as fIn:
      publicKey = minisign.PublicKey.from_bytes(fIn.read())
    try:
      publicKey.verify(metadataData, signature)
      if 'pubkey_url' in signature.trusted_comment:
        path = json.loads(signature.trusted_comment)['pubkey_url']
        response = requests.get(path, timeout=10)
        if response.ok:
          pubKeyRemote = response.content.strip()
          if pubKeyRemote.decode() == str(bytes(publicKey))[2:-1]:
            logging.info('Success: remote and local key match')
          else:
            logging.error('remote and local key differ', exc_info=True)
            raise minisign.VerifyError
      print('Signature is acceptable:\n', json.dumps(json.loads(signature.trusted_comment), indent=2))
      return True
    except minisign.VerifyError:
      logging.error('VERIFICATION', exc_info=True)
    return False



""" Custom tree view on data model """
import logging
import os
from enum import Enum
from pathlib import Path
from typing import Any
from PySide6.QtCore import Qt
from PySide6.QtGui import QDropEvent, QEventPoint, QStandardItem, QStandardItemModel
from PySide6.QtWidgets import QAbstractItemView, QMenu, QMessageBox, QTreeView, QWidget
from ..backendWorker.worker import Task
from ..miscTools import callAddOn
from .guiCommunicate import Communicate
from .guiStyle import Action
from .messageDialog import showMessage
from .projectLeafRenderer import ProjectLeafRenderer


class TreeView(QTreeView):
  """ Custom tree view on data model """
  def __init__(self, parent:QWidget, comm:Communicate, model:QStandardItemModel):
    super().__init__(parent)
    self.aParentWidget = parent
    self.comm = comm
    self.setModel(model)
    self.setHeaderHidden(True)
    self.setStyleSheet('QTreeView::branch {border-image: none;}')
    self.setIndentation(40)
    self.renderer = ProjectLeafRenderer(self.comm)
    self.setItemDelegate(self.renderer)
    self.setExpandsOnDoubleClick(False)
    self.setAcceptDrops(True)
    self.setDropIndicatorShown(True)
    self.setDefaultDropAction(Qt.DropAction.MoveAction)
    self.setDragDropMode(QAbstractItemView.DragDropMode.InternalMove)
    self.doubleClicked.connect(self.tree2Clicked)


  def contextMenuEvent(self, p:QEventPoint) -> None:                                  # type: ignore[override]
    """
    create context menu

    Args:
      p (QPoint): point of clicking
    """
    item = self.model().itemFromIndex(self.currentIndex())                        # type: ignore[attr-defined]
    if item is None:                                                                 #clicked outside any leaf
      return
    folder = item.data()['hierStack'].split('/')[-1][0]=='x'
    context = QMenu(self)
    if folder:
      Action('Add child folder',                   self, [Command.ADD_CHILD],      context)
    Action('Add sibling folder',                   self, [Command.ADD_SIBLING],    context)
    Action('Delete item',                          self, [Command.DELETE],         context)
    context.addSeparator()
    Action('Hide/show item details',               self, [Command.SHOW_DETAILS], context)
    Action('Mark item as hidden/shown',            self, [Command.HIDE],           context)
    context.addSeparator()
    if not folder:
      Action('Open file with another application', self, [Command.OPEN_EXTERNAL],    context)
    Action('Open folder in file browser',          self, [Command.OPEN_FILEBROWSER], context)
    if folder:
      projectGroup = self.comm.configuration['projectGroups'][self.comm.projectGroup]
      if projectAddOns := projectGroup.get('addOns',{}).get('project',''):
        context.addSeparator()
        for label, description in projectAddOns.items():
          Action(description, self, [Command.ADD_ON, label], context)
    context.exec(p.globalPos())                                                   # type: ignore[attr-defined]
    return


  def execute(self, command:list[Any]) -> None:
    """
    after selecting a item from context menu

    Args:
      command (list): list of commands
    """
    item = self.model().itemFromIndex(self.currentIndex())                        # type: ignore[attr-defined]
    hierStack = item.data()['hierStack'].split('/')
    if command[0] is Command.ADD_CHILD:
      self.comm.uiRequestTask.emit(Task.ADD_DOC, {'hierStack':hierStack, 'docType':'x1', 'doc':{'name':'new item'}})

    elif command[0] is Command.ADD_SIBLING:
      hierStack= hierStack[:-1]
      self.comm.uiRequestTask.emit(Task.ADD_DOC, {'hierStack':hierStack, 'docType':'x1', 'doc':{'name':'new item'}})

    elif command[0] is Command.DELETE:
      ret = QMessageBox.critical(self, 'Warning', 'Are you sure you want to delete this data?',\
                                 QMessageBox.StandardButton.Yes, QMessageBox.StandardButton.No)
      if ret==QMessageBox.StandardButton.Yes:
        docID = hierStack[-1]
        self.comm.uiRequestTask.emit(Task.DELETE_DOC, {'docID':docID})
        self.comm.changeProject.emit(self.parent().projID, '')                    # type: ignore[attr-defined]
        return

    elif command[0] is Command.SHOW_DETAILS:
      gui    = item.data()['gui']
      gui[0] = not gui[0]
      docID  = hierStack[-1]
      def iterate(currentItem:QStandardItem) -> None:
        """ iterate through all branches and leaves and find items matching the docID
        Args:
          currentItem (QStandardItem): item to iterate to its children
        """
        currentIndex = self.model().indexFromItem(currentItem)                    # type: ignore[attr-defined]
        if currentItem.data() is not None and docID==currentItem.data()['hierStack'].split('/')[-1]:
          currentItem.setData({ **currentItem.data(), **{'gui':gui}})
        for row in range(self.model().rowCount(currentIndex)):
          for column in range(self.model().columnCount(currentIndex)):
            childIndex = self.model().index(row, column, currentIndex)
            iterate(self.model().itemFromIndex(childIndex))                       # type: ignore[attr-defined]
      iterate(self.model().invisibleRootItem())                                   # type: ignore[attr-defined]
      # only one change once the DB
      self.comm.uiRequestTask.emit(Task.SET_GUI, {'docID':docID, 'gui':gui})

    elif command[0] is Command.HIDE:
      logging.debug('hide document %s',hierStack[-1])
      self.comm.uiRequestTask.emit(Task.HIDE_SHOW, {'docID':hierStack[-1]})
      #TODO: current implementation: you hide one; all others are hidden as well
      # Talk to GW what is the default expectation; system allows for individual hiding
      #
      # after hide, do not hide immediately but wait on next refresh

    elif command[0] is Command.OPEN_EXTERNAL or command[0] is Command.OPEN_FILEBROWSER:
      # depending if non-folder / folder; address different item in hierstack
      docID = hierStack[-2] \
        if command[0] is Command.OPEN_FILEBROWSER and hierStack[-1][0]!='x' \
        else hierStack[-1]
      self.comm.uiRequestTask.emit(Task.OPEN_EXTERNAL, {'docID':docID})

    elif command[0] is Command.ADD_ON:
      callAddOn(command[1], self.comm, item.data()['hierStack'], self)
    else:
      logging.error('Unknown context menu %s', command, exc_info=True)
    self.comm.uiRequestHierarchy.emit(self.parent().projID, self.parent().showAll)# type: ignore[attr-defined]
    return


  def scrollToDoc(self, docID:str) -> None:
    """
    Scroll to document with docID

    Args:
      docID (str): document ID
    """
    def iterate(currentItem:QStandardItem) -> QStandardItem | None:
      """ iterate through all branches and leaves and find items matching the docID
      Args:
        currentItem (QStandardItem): item to iterate to its children
      Returns:
        QStandardItem | None: item with docID or None if not found
      """
      currentIndex = self.model().indexFromItem(currentItem)                      # type: ignore[attr-defined]
      if currentItem.data() is not None and docID==currentItem.data()['hierStack'].split('/')[-1]:
        return currentItem
      for row in range(self.model().rowCount(currentIndex)):
        for column in range(self.model().columnCount(currentIndex)):
          childIndex = self.model().index(row, column, currentIndex)
          found = iterate(self.model().itemFromIndex(childIndex))                 # type: ignore[attr-defined]
          if found is not None:
            return found
      return None
    item = iterate(self.model().invisibleRootItem())                              # type: ignore[attr-defined]
    if item is not None:
      # TODO scroll does not work even if item is not none and visible
      # there is some fast scrolling to some locatino, but then the view scrolls back
      self.scrollTo(item.index(), QAbstractItemView.EnsureVisible)                # type: ignore[attr-defined]
    return


  def tree2Clicked(self) -> None:
    """
    after double-click on tree leaf: open form
    - no redraw required since renderer asks automatically for update
    """
    item = self.model().itemFromIndex(self.currentIndex())                        # type: ignore[attr-defined]
    docID = item.data()['hierStack'].split('/')[-1]
    self.comm.formDoc.emit({'id':docID})
    return


  def dragEnterEvent(self, event:QDropEvent) -> None:
    """
    Override default: what happens if you drag an item

    Args:
      event (QMouseEvent): event
    """
    event.acceptProposedAction()
    return


  def dropEvent(self, event:QDropEvent) -> None:
    """
    Override default: what happens at end of drag&drop

    Args:
      event (QDropEvent): event
    """
    if event.mimeData().hasUrls():                                                    #file dropped onto pasta
      item = self.model().itemFromIndex(self.indexAt(event.pos()))                # type: ignore[attr-defined]
      if item is None or item.data()['docType'][0][0]!='x':
        showMessage(self, 'Error', 'You can drop external files only onto folders.')
        return
      # create a list of all files
      files, folders = [], []
      for url in event.mimeData().urls():
        path = url.toLocalFile()
        if Path(path).is_file():
          files.append(path)
        else:
          files +=   list(Path(path).rglob('*'))                                      # type: ignore[arg-type]
          folders += [x[0] for x in os.walk(path)]
      if not (folders+[str(i) for i in files]):
        showMessage(self, 'Error', 'The files seem empty.')
        return
      docID = item.data()['hierStack'].split('/')[-1]
      self.comm.uiRequestTask.emit(Task.DROP_EXTERNAL, {'docID':docID, 'files':files, 'folders':folders})
      event.ignore()
    elif 'application/x-qstandarditemmodeldatalist' in event.mimeData().formats():
      super().dropEvent(event)
    else:
      logging.error('Drop unknown data: %s', event.mimeData().formats(), exc_info=True)
    return


class Command(Enum):
  """ Commands used in this file """
  ADD_CHILD        = 1
  ADD_SIBLING      = 2
  DELETE           = 3
  SHOW_DETAILS     = 4
  HIDE             = 5
  OPEN_EXTERNAL    = 6
  OPEN_FILEBROWSER = 7
  ADD_ON           = 8



""" Misc functions that do not require instances """
import importlib
import json
import logging
import os
import platform
import socket
import subprocess
import sys
import tempfile
import time
from collections.abc import Mapping
from pathlib import Path
from typing import Any, Union
from urllib import request
import pandas as pd
import requests
from anytree import Node
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
from packaging.version import parse as parse_version
from PySide6.QtCore import Slot
from PySide6.QtWidgets import QWidget
import pasta_eln
from .fixedStringsJson import CONF_FILE_NAME, configurationGUI, defaultConfiguration


class Bcolors:
  """
  Colors for Command-Line-Interface and output
  """
  if platform.system()=='Windows':
    HEADER = ''
    OKBLUE = ''
    OKGREEN = ''
    WARNING = ''
    FAIL = ''
    ENDC = ''
    BOLD = ''
    UNDERLINE = ''
  else:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

def updateAddOnList(projectGroup:str='') -> dict[str, Any]:
  """
  Rules:
  - each data-type in its own try-except
  - inside try: raise ValueError exception on failure/None
  - except empty: pass
  - all descriptions in type have to be small letters
  - if want to force to skip top datatypes and use one at bottom: if doctype... -> exception

  Args:
    projectGroup (str): project group to scan

  Returns:
    dict: dict with all add-ons
  """
  with open(Path.home()/CONF_FILE_NAME, encoding='utf-8') as f:
    configuration = json.load(f)
  if not projectGroup:
    projectGroup = configuration['defaultProjectGroup']
  directory = Path(configuration['projectGroups'][projectGroup]['addOnDir'])
  sys.path.append(str(directory))                                                               #allow add-ons
  # Add-Ons
  verboseDebug = False
  extractorsAll= {}
  otherAddOns:dict[str,dict[str,str]]  = {'project':{}, 'table':{} ,'definition':{} ,'form':{}, '_ERRORS_':{}}
  for fileName in os.listdir(directory):
    # Extractor
    if fileName.endswith('.py') and fileName.startswith('extractor_'):
      #start with file
      with open(directory/fileName, encoding='utf-8') as fIn:
        if verboseDebug: print('\n'+fileName)
        lines = fIn.readlines()
        extractorsThis = {}
        ifInFile, headerState, header = False, True, []
        for idx,line in enumerate(lines):
          line = line.rstrip()
          if idx>0 and '"""' in line:
            headerState = False
          if headerState:
            line = line.replace('"""','')
            header.append(line)
            continue
          if 'if' in line and '#:' in line:
            if verboseDebug: print('line', line)
            specialType = line.split('==')[1].split(':')[0].strip(" '"+'"')
            if verboseDebug: print('  special',specialType)
            extractorsThis[specialType] = line.split('#:')[1].strip()
            ifInFile = True
          elif 'else:' in line and '#:' in line:
            otherAddOns['_ERRORS_'][fileName] = 'ERROR there should not be an else in the code'
          elif 'return' in line and 'recipe' in line and not ifInFile:
            if verboseDebug: print('line', line)
            if line.count('recipe')==1:
              linePart = line.split('recipe')[1].strip()
              linePart = linePart.split(':')[1].split(',')[0].strip(" '"+'"')
            elif line.count('recipe')==2:
              possLines = [i.strip() for i in lines if ('recipe' in i and '=' in i and 'def' not in i)]
              if len(possLines)==1:
                linePart=possLines[0].split('=')[1].strip(" '"+'"')
              elif len(possLines)>1:
                logging.warning('Could not decipher %s. Take shortest!', fileName)
                linePart=sorted(possLines)[0].split('=')[1].strip(" '"+'"')
              else:
                otherAddOns['_ERRORS_'][fileName] = 'ERROR could not decipher code'
                linePart=''
            extractorsThis[linePart]='Default'
            if verboseDebug:
              print('  return', linePart)
        if verboseDebug:
          print('Extractors', extractorsThis)
        ending = fileName.split('_')[1].split('.')[0]
        extractorsAll[ending]=extractorsThis
    # header not used for now
    # Project, et al
    if fileName.endswith('.py') and '_' in fileName and fileName.split('_')[0] in ['project','table','definition','form']:
      name        = fileName[:-3]
      try:
        module      = importlib.import_module(name)
        description = module.description
        _ = module.reqParameter                                                 # check if reqParameter exists
        otherAddOns[fileName.split('_')[0]][name] = description
      except Exception as e:
        description = f'** SYNTAX ERROR in add-on **: {e}'
        otherAddOns['_ERRORS_'][name] = description
  #update configuration file
  configuration['projectGroups'][projectGroup]['addOns']['extractors'] = extractorsAll
  configuration['projectGroups'][projectGroup]['addOns'] |= otherAddOns
  with open(Path.home()/CONF_FILE_NAME,'w', encoding='utf-8') as f:
    f.write(json.dumps(configuration, indent=2))
  return {'addon directory':directory} | extractorsAll | otherAddOns


def installPythonPackages(directory:str) -> None:
  """Install a Python packages using pip depending on files in add-on folder
  Args:
    directory (str): path to the add-on folder
  """
  allLibs = set()
  for fileName in os.listdir(directory):
    if fileName.endswith('.py'):
      #start with file
      with open(Path(directory)/fileName, encoding='utf-8') as fIn:
        lines = [i.strip() for i in fIn.readlines() if 'import' in i]
        libsList =  [i.split('import')[1].strip().split(',') for i in lines if i.startswith('import')]
        libs =  [i.strip() for j in libsList for i in j]
        libs += [i.split('from')[1].strip().split('import')[0].strip() for i in lines if i.startswith('from')]
        libs = [i.split('.')[0] for i in libs]                                           #only look at package
        libs = [i.split()[0] for i in libs]                                  #get rid of all things at the end
        allLibs.update(libs)
  allLibs = allLibs.difference(sys.stdlib_module_names)        # all libs that are not in the standard library
  allLibs = allLibs.difference([i[:-3] for i in os.listdir(directory) if i.endswith('.py')])#remove all libs that are in the directory
  allLibs = allLibs.difference({i.split('.')[0] for i in sys.modules})# remove libs used by pasta, already in use
  for lib in allLibs:
    try:
      importlib.import_module(lib)                                 # check if the package is already installed
    except ImportError:
      if lib == 'sklearn':
        lib = 'scikit-learn'
      subprocess.check_call([sys.executable, '-m', 'pip', 'install', lib])
  return


def callAddOn(name:str, comm:Any, content:Any, widget:QWidget) -> Any:
  """ Call add-ons
  Args:
    name (str): name of the add-on
    comm (Any): communication object
    content (Any): content to be consumed by addon, e.g. projectID, document
    widget: widget to be used

  Returns:
    Any: result of the add-on
  """
  module      = importlib.import_module(name)
  parameter   = comm.configuration.get('addOnParameter', {})
  try:
    subParameter = parameter[name]
  except KeyError:
    print('**Info: No parameter for this add-on')
    subParameter = {}
  return module.main(comm, content, widget, subParameter)


def callDataExtractor(docID:str, comm:Any) -> Any:
  """ Call data extractor for a given file path: CALL THE DATA FUNCTION

  Args:
    docID (str): docID
    comm (Communication): communication layer

  Returns:
    Any: result of the data extractor
  """
  doc = getDoc(comm, docID)
  filePath = comm.basePath/doc['branch'][0]['path']
  if not isinstance(filePath, Path):
    filePath = Path(filePath)
  extension = filePath.suffix[1:]                                                   #cut off initial . of .jpg
  if str(filePath).startswith('http'):
    absFilePath = Path(tempfile.gettempdir())/filePath.name
    with request.urlopen(filePath.as_posix().replace(':/','://'), timeout=60) as urlRequest:
      with open(absFilePath, 'wb') as f:
        try:
          f.write(urlRequest.read())
        except Exception:
          print('Error saving downloaded file to temporary disk')
  else:
    if filePath.is_absolute():
      filePath  = filePath.relative_to(comm.basePath)
    absFilePath = comm.basePath/filePath
  pyFile = f'extractor_{extension.lower()}.py'
  pyPath = Path(comm.addOnPath)/pyFile
  if pyPath.is_file():
    # import module and use to get data
    try:
      module = importlib.import_module(pyFile[:-3])
      return module.data(absFilePath, {})
    except Exception as e:
      logging.warning('CallDataExtractor: %s',e)
  return None


def getHierarchy(comm:Any, docID:str, showAll:bool=True) -> tuple[Node, dict[str, Any]]:
  """ Helper for add-ons: get hierarchy of a project from backend
  Args:
    comm (Communicate): communicate-backend
    docID (str): project ID
    showAll (bool): show all nodes, even hidden ones

  Returns:
    tuple: (hierarchy as anytree, project document as dict)
  """
  hierarchyI = None
  projDoc   = None
  @Slot(Node, dict)
  def receiveData(h:Node, doc:dict[str,Any]) -> None:
    nonlocal hierarchyI
    hierarchyI = h
    nonlocal projDoc
    projDoc   = doc
  comm.backendThread.worker.beSendHierarchy.connect(receiveData)
  comm.uiRequestHierarchy.emit(docID, showAll)
  while hierarchyI is None or projDoc is None:
    time.sleep(0.1)
  return hierarchyI, projDoc


def getDoc(comm:Any, docID:str) -> dict[str, Any]:
  """ Helper for add-ons: get document from backend

  Args:
    comm (Communicate): communicate-backend
    docID (str): document ID

  Returns:
    dict: document as dict
  """
  doc = None
  @Slot(dict)
  def receiveData(iDoc:dict[str,Any]) -> None:
    """ Slot to receive data
    Args:
      iDoc (dict): document
    """
    nonlocal doc
    doc = iDoc
  comm.backendThread.worker.beSendDoc.connect(receiveData)
  comm.uiRequestDoc.emit(docID)
  while doc is None:
    time.sleep(0.1)
  return doc


def isFloat(val:str) -> bool:
  """Check if a value can be converted to float
  Args:
    val (str): value to check
  Returns:
    bool: True if value can be converted to float, False otherwise
  """
  try:
    float(val)
    return True
  except (ValueError, TypeError):
    return False


def dfConvertColumns(df:pd.DataFrame, ratio:int=10) -> pd.DataFrame:
  """Convert columns in a DataFrame to numeric if a significant
  portion of the column can be converted

  Args:
    df (pd.DataFrame): DataFrame with columns to convert
    ratio (int): threshold ratio to determine if a column should be converted

  Returns:
    pd.DataFrame: DataFrame with converted columns
  """
  for col in df.columns:
    numConvertible = df[col].apply(isFloat).sum()
    if numConvertible > len(df) / ratio:
      df[col] = pd.to_numeric(df[col], errors='coerce')
  return df


class MplCanvas(FigureCanvas):
  """ Canvas to draw upon """
  def __init__(self, _:Any=None, width:float=5, height:float=4, dpi:int=100):
    """
    Args:
      _ (Any): figure
      width (float): width in inch
      height (float): height in inch
      dpi (int): dots per inch
    """
    self.figure = Figure(figsize=(width, height), dpi=dpi)
    self.axes   = self.figure.add_subplot(111)
    super().__init__(self.figure)


def getConfiguration(defaultProjectGroup:str='') -> tuple[dict[str, Any],str]:
  """ Get configuration from home directory
  Args:
    defaultProjectGroup (str): project group to use, if not given, use default
  Returns:
    tuple: configuration dict and default project group
  """
  configuration = defaultConfiguration
  configFileName = Path.home() / CONF_FILE_NAME
  if configFileName.is_file():
    with open(configFileName, encoding='utf-8') as confFile:
      configuration |= json.load(confFile)
  for _, items in configurationGUI.items():
    for k,v in items.items():
      if k not in configuration['GUI']:
        configuration['GUI'][k] = v[1]
  if configuration['version'] != 3:
    print('**Info: configuration file does not exist or version is != 3')
    return {},''
  defaultProjectGroup = defaultProjectGroup or configuration['defaultProjectGroup']
  if defaultProjectGroup not in configuration['projectGroups']:
    listPossible = list(configuration['projectGroups'].keys())
    raise ValueError(f'BadProjectGroup: {defaultProjectGroup}. Possible: {listPossible}')
  return configuration, defaultProjectGroup


def hardRestart() -> None:
  """
  Complete restart: cold restart
  """
  try:
    os.execv('pastaELN',[''])                                                               #installed version
  except Exception:
    os.execv(sys.executable, [sys.executable,'-m','pasta_eln.gui'])      #started for programming or debugging
  return


def isConnectedToInternet() -> bool:
  """Check if the system is connected to the internet

  Returns:
    bool: True if connected, False otherwise
  """
  try:
    socket.setdefaulttimeout(3)
    socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect(('8.8.8.8', 53))
    return True
  except OSError:
    pass
  return False


def getRORIDLabel(idString:str) -> str :
  """ Get label from RORID
  Args:
    idString (str): RORID
  Returns:
    str: label of the RORID
  """
  reply = requests.get(f'https://api.ror.org/organizations/{idString}', timeout=10)
  labels = [i for i in reply.json()['names'] if 'ror_display' in i['types']]
  labels = labels or [i for i in reply.json()['names'] if 'acronym' not in i['types']]
  labels = labels or reply.json()['names']
  return labels[0]['value'] if labels else ''


def getORCIDName(idString:str) -> tuple[str, str]:
  """ Get name from ORCID
  Args:
    idString (str): ORCID
  Returns:
    tuple: (first name, last name)
  """
  reply = requests.get(f'https://pub.orcid.org/v3.0/{idString}', timeout=10)
  text = reply.content.decode()
  first = text.split('<personal-details:given-names>')[1].split('</personal-details:given-names>')[0]
  last = text.split('<personal-details:family-name>')[1].split('</personal-details:family-name>')[0]
  return first, last


def testNewPastaVersion(update:bool=False) -> bool:
  """ Test if this version is up to date with the latest version on pypi
  - variable largestVersionOnPypi is the latest NON-BETA version on pypi

  Args:
    update (bool): update to latest version

  Returns:
    bool: if up-to-date or if current version is a beta
  """
  if not isConnectedToInternet():
    return True
  if update:
    subprocess.check_call([sys.executable, '-m', 'pip', 'install', '--upgrade', 'pasta-eln'])
    hardRestart()
  url = 'https://pypi.org/pypi/pasta-eln/json'
  with request.urlopen(url) as response:
    data = json.loads(response.read())
  releases = [i for i in list(data['releases'].keys()) if 'b' not in i]                 # remove beta versions
  largestVersionOnPypi = sorted(releases, key=parse_version)[-1]
  return largestVersionOnPypi == pasta_eln.__version__ or 'b' in pasta_eln.__version__


# adapted from flatten-dict https://github.com/ianlini/flatten-dict
# - reduce dependencies and only have python 3 code
# - add conversion of dict to list if applicable
def flatten(d:dict[Any,Any], keepPastaStruct:bool=False) -> dict[object, Any]:
  """Flatten `Mapping` object

  Args:
    d : dict-like object
        The dict that will be flattened
    keepPastaStruct : bool
        keep pasta elements from flattening

  Returns:
    flat_dict : dict
  """
  enumerate_types=(list,)
  flatten_types = (Mapping,) + enumerate_types
  flat_dict = {}

  def dot_reducer(k1:object, k2:object) -> Union[str, object]:
    """ Reducer function """
    return k2 if k1 is None else f"{k1}.{k2}"

  def _flatten(_d:Union[Mapping[Any, Any], list[Any]], depth:int, parent:object=None) -> bool:
    """ Recursive function """
    key_value_iterable = enumerate(_d) if isinstance(_d, enumerate_types) else _d.items()
    has_item = False
    for key, value in key_value_iterable:
      has_item = True
      flat_key = dot_reducer(parent, key)
      if isinstance(value, flatten_types):
        # recursively build the result
        has_child = _flatten(value, depth=depth + 1, parent=flat_key)
        if has_child or not isinstance(value, ()):# ignore key in this level because it already has child key or its value is empty
          continue
      if flat_key in flat_dict and parent is not None:
        continue  # skip if key already exists and if data comes from database after merging with changed data
      # add an item to the result
      flat_dict[flat_key] = value
    return has_item

  # start recursive calling
  backup = {'type':d.pop('type',None), 'branch':d.pop('branch',None),   'tags':d.pop('tags',None),
            'gui':d.pop('gui',None),   'qrCodes':d.pop('qrCodes',None), '_ids':d.pop('_ids',None)} \
           if keepPastaStruct else {}
  _flatten(d, depth=1)
  return flat_dict | {k:v for k,v in backup.items() if v is not None}


def hierarchy(d:dict[str,Any]) -> dict[str,Any]:
  """Reverse flattening of dict-like object

  Args:
    d : dict-like object
      The dict that will be reversed

  Returns
    normalDict : dict
  """
  def dot_splitter(flat_key:str) -> tuple[str, ...]:
    """ split using the . symbol """
    keys = tuple(flat_key.split('.'))
    return keys

  def nested_set_dict(d:dict[str,Any], keys:tuple[str, ...], value:Any) -> None:
    """Set a value to a sequence of nested keys

    Args:
      d : Mapping
      keys : Sequence[str]
      value : Any
    """
    key = keys[0]
    if len(keys) == 1:
      if key in d:
        raise ValueError(f"duplicated key '{key}'")
      try:
        d[key] = value
      except TypeError as e:
        print('TypeError:', type(d), d, type(key), key, type(value), value)
        raise e
      return
    d = d.setdefault(key, {})
    nested_set_dict(d, keys[1:], value)
    return

  def dict2list(d:dict[str,Any]) -> Union[list[dict[str,Any]], dict[str,Any]]:
    """ convert a dictionary to list if all keys are numbers """
    for key in d:
      if isinstance(d[key], dict):
        d[key] = dict2list(d[key])
    if all(i.isdigit() for i in d):
      d = list(d.values())                                                          # type: ignore[assignment]
    return d

  # start recursion
  normalDict:dict[str,Any] = {}
  for flat_key, value in d.items():
    key_tuple = dot_splitter(flat_key)
    nested_set_dict(normalDict, key_tuple, value)
  normalDict =  dict2list(normalDict)                                               # type: ignore[assignment]
  return normalDict



""" Configuration for terminology lookup services. """
from typing import Any

lookupConfig:list[dict[str, Any]] = []


# Wikipedia configuration
# requests.get('https://en.wikipedia.org/w/rest.php/v1/search/page',
#              params={'q': 'force', 'limit': 5},
#              headers={'User-Agent': 'PASTA-ELN (https://github.com/PASTA-ELN/pasta-eln)'},
#              timeout=10)
lookupConfig.append({
    'name': 'wikipedia',
    'url': 'https://en.wikipedia.org/w/rest.php/v1/search/page',
    'search_term_key': 'q',
    'request_params': {
      'q': 'searchTerm',
      'limit': 5
    },
    'header': {
      'User-Agent': 'PASTA-ELN (https://github.com/PASTA-ELN/pasta-eln)'
    },
    'iri_prefix': 'https://en.wikipedia.org/w/index.php?curid=',
    'search_criteria': {
      'results_keys': [
        'pages'
      ],
      'description_keys': [
        'description'
      ],
      'id_key': 'id'
    },
    'skip_description': 'Topics referred to by the same term',
    'icon_name': 'wikipedia.png'
  })


# Wikidata configuration
# requests.get('https://www.wikidata.org/w/api.php',
#              params={'search': 'force', 'action': 'wbsearchentities', 'format': 'json', 'language': 'en', 'type': 'item', 'continue': '0'},
#              headers={'User-Agent': 'PASTA-ELN (https://github.com/PASTA-ELN/pasta-eln)'},
#              timeout=10)
lookupConfig.append({
    'name': 'wikidata',
    'url': 'https://www.wikidata.org/w/api.php',
    'search_term_key': 'search',
    'request_params': {
      'search': 'searchTerm',
      'action': 'wbsearchentities',
      'format': 'json',
      'language': 'en',
      'type': 'item',
      'continue': '0'
    },
    'header': {
      'User-Agent': 'PASTA-ELN (https://github.com/PASTA-ELN/pasta-eln)'
    },
    'search_criteria': {
      'results_keys': [
        'search'
      ],
      'description_keys': [
        'display',
        'description',
        'value'
      ],
      'id_key': 'concepturi'
    },
    'icon_name': 'wikidata.png'
  })


# Ontology Lookup Service (OLS) configuration
# requests.get('http://www.ebi.ac.uk/ols/api/search',
#              params={'q': 'force'},
#              headers={},
#              timeout=10)
lookupConfig.append({
    'name': 'ontology_lookup_service',
    'url': 'http://www.ebi.ac.uk/ols/api/search',
    'search_term_key': 'q',
    'request_params': {
      'q': 'searchTerm'
    },
    'search_criteria': {
      'results_keys': [
        'response',
        'docs'
      ],
      'description_keys': [
        'description'
      ],
      'id_key': 'iri'
    },
    'icon_name': 'ols.png'
  })


# TIB Terminology Service configuration
# requests.get('https://service.tib.eu/ts4tib/api/search',
#              params={'q': 'force'},
#              headers={'Caller': 'PASTA_ELN'},
#              timeout=10)
lookupConfig.append({
    'name': 'tib_terminology_service',
    'url': 'https://service.tib.eu/ts4tib/api/search',
    'search_term_key': 'q',
    'request_params': {
      'q': 'searchTerm'
    },
    'header': {
      'Caller': 'PASTA_ELN'
    },
    'duplicate_ontology_names': [
      'afo', 'bco', 'bto', 'chiro', 'chmo', 'duo', 'edam', 'efo', 'fix', 'hp', 'iao', 'mod', 'mop', 'ms',
      'nmrcv', 'ncit', 'obi', 'om', 'pato', 'po', 'proco', 'prov', 'rex', 'ro', 'rxno', 'sbo', 'sepio', 'sio',
      'swo', 't4fs', 'uo'
    ],
    'search_criteria': {
      'results_keys': [
        'response',
        'docs'
      ],
      'description_keys': [
        'description'
      ],
      'id_key': 'iri',
      'ontology_name_key': 'ontology_name'
    },
    'icon_name': 'tib.png'
  })



import logging
from pasta_eln.UI.guiCommunicate import Communicate
from pasta_eln.UI.config.main import Configuration

def test_simple(qtbot, caplog):

  comm = Communicate('research')
  while comm.backendThread.worker.backend is None:
    qtbot.wait(100)

  dialog = Configuration(comm)
  dialog.setMinimumSize(1024,800)
  dialog.show()
  qtbot.addWidget(dialog)
  qtbot.wait(1000)
  path = qtbot.screenshot(dialog)
  print(path)

  comm.shutdownBackendThread()

  errors = [record for record in caplog.records if record.levelno >= logging.ERROR]
  assert not errors, f"Logging errors found: {[record.getMessage() for record in errors]}"
